"""
Tools to plot that work together with the likelihood.

Warning:
------------
This is NOT a comprehensive plotting library. Instead, the functions below are largely designed as
templates that you should modify for your specific purposes.

Overview:
------------------------

Helper for pretty LaTeX parameter labels and general plot style:

* :const:`PAR_LABEL_DICT`
* :func:`set_plot_style`

Showing the parameter covariance matrix:

* :func:`triangle_plot_from_cov`
* :func:`add_confidence_ellipse`

Analysing minimisers:

* :func:`add_min_trajectory`
* :func:`plot_minimiser_convergence`

Other:

* :func:`plot_band_powers`
* :func:`plot_mcmc_chain_steps`
* :func:`plot_foreground_components`
"""

# --------------------------------------#
# IMPORTS
# --------------------------------------#

from candl.lib import *

# --------------------------------------#
# CONSTANTS AND HELPERS
# --------------------------------------#

PAR_LABEL_DICT = {
    "ombh2": "\\Omega_{\\mathrm{b}}h^2",
    "omch2": "\\Omega_{\\mathrm{c}}h^2",
    "theta": "100\\theta_{\\mathrm{MC}}",
    "clamp": "10^9 A_{\\mathrm{s}} e^{-2\\tau}",
    "logA": "\\log{ A_{\\mathrm{s}}}",
    "ns": "n_{\\mathrm{s}}",
    "H0": "H_{\\mathrm{0}}",
    "sigma8": "\\sigma_8",
    "tau": "\\tau",
    "S8": "S_8 \\equiv \\sigma_8 \\sqrt{\\Omega_{\\mathrm{m}}/0.3}",
    "omegal": "\\Omega_{\\mathrm{\\Lambda}}",
    "age": "{\\mathrm{Age}}/{\\mathrm{Gyr}}",
    "omega_b": "\\Omega_{\\mathrm{b}}h^2",
    "omega_cdm": "\\Omega_{\\mathrm{c}}h^2",
    "ln10^{10}A_s": "\\log{ A_{\\mathrm{s}}}",
    "n_s": "n_{\\mathrm{s}}",
    "tau_reio": "\\tau",
}
"""
Dictionary of common names of and abbreviations for cosmological parameters and corresponding pretty latex strings.
"""


# --------------------------------------#
# PLOT STYLE
# --------------------------------------#


def set_plot_style():
    """
    Sets the plotting style. Important to unify things across figures and machines and generally makes plots prettier.
    Thank you to Federico Bianchini for this template!
    """
    # rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
    # plt.rcParams.update({
    # "text.usetex": True,
    # "font.family": "serif"
    # })
    rc("text", usetex=True)
    # rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
    # plt.rcParams['font.family'] = 'sans-serif'
    # plt.rcParams['font.sans-serif'] = ['Tahoma']
    # plt.rcParams['mathtext.fontset'] = 'cm'
    # rc('font',**{'family':'sans-serif'})#,'dejavuserif':['Computer Modern']})
    # plt.rcParams["mathtext.fontset"] = "dejavuserif"
    plt.rcParams["axes.linewidth"] = 1.5
    plt.rcParams["axes.labelsize"] = 10
    plt.rcParams["axes.titlesize"] = 12
    plt.rcParams["xtick.labelsize"] = 8  # 12
    plt.rcParams["ytick.labelsize"] = 8  # 12
    plt.rcParams["xtick.major.size"] = 4  # 7
    plt.rcParams["ytick.major.size"] = 4  # 7
    plt.rcParams["xtick.minor.size"] = 2  # 4
    plt.rcParams["ytick.minor.size"] = 2  # 4
    plt.rcParams["legend.fontsize"] = 10
    plt.rcParams["legend.frameon"] = False

    plt.rcParams["xtick.major.width"] = 1
    plt.rcParams["ytick.major.width"] = 1
    plt.rcParams["xtick.minor.width"] = 1
    plt.rcParams["ytick.minor.width"] = 1
    plt.clf()
    plt.close()
    # sns.set(rc('font',**{'family':'serif','serif':['Computer Modern']}))
    # sns.set_style("ticks", {'figure.facecolor': 'grey'})


# --------------------------------------#
# CONFIDENCE ELLIPSE
# --------------------------------------#


def add_confidence_ellipse(
    ax, this_cov, ix=0, jx=1, mean_x=0, mean_y=0, n_std=1, facecolor="none", **kwargs
):
    """
    Add a confidence based on the covariance matrix to ax.
    Based on matplotlib example code with some slight modifications.
    See: https://matplotlib.org/stable/gallery/statistics/confidence_ellipse.html

    There may be a slight mix-up between ix/jx and mean_x/mean_y, worth trying out some permutations if you use
    this function by itself.

    Parameters
    --------------
    ax : matplotlib.axes
        The axes to add the ellipse to.
    this_cov : array
        Parameter covariance matrix.
    ix : int
        Index of first parameter.
    jx : int
        Index of first parameter.
    mean_x : float
        Where to centre the ellipse in x.
    mean_y : float
        Where to centre the ellipse in y.
    n_std : float
        Which number of standard deviations to show.
    facecolor : string/array
        Facecolour of ellipse.
    kwargs :
        Other arguments to be passed through to matplotlib.patches.Ellipse.
    """

    pearson = this_cov[ix, jx] / np.sqrt(this_cov[ix, ix] * this_cov[jx, jx])
    ell_radius_x = np.sqrt(1 + pearson)
    ell_radius_y = np.sqrt(1 - pearson)
    ellipse = Ellipse(
        (0, 0),
        width=ell_radius_x * 2,
        height=ell_radius_y * 2,
        facecolor=facecolor,
        **kwargs,
    )

    # Important scaling for 2 dimensions - otherwise contours are too tight.
    # See information here https://arxiv.org/pdf/0906.4123.pdf
    prob_amount = sp.stats.norm.cdf(n_std) - sp.stats.norm.cdf(-n_std)
    scale_factor = np.sqrt(sp.stats.chi2.ppf(prob_amount, 2))

    scale_x = np.sqrt(this_cov[ix, ix]) * scale_factor
    scale_y = np.sqrt(this_cov[jx, jx]) * scale_factor

    transf = (
        transforms.Affine2D()
        .rotate_deg(45)
        .scale(scale_x, scale_y)
        .translate(mean_x, mean_y)
    )

    ellipse.set_transform(transf + ax.transData)

    return ax.add_patch(ellipse)


# --------------------------------------#
# TRIANGLE PLOT WITH CONTOURS
# --------------------------------------#


def triangle_plot_from_cov(
    pars_to_plot,
    bf_point,
    par_cov,
    pars_in_cov,
    sigma_plot_range=5,
    sigma_levels=2,
    ax=None,
    contour_colours=None,
    contour_edgecolour=None,
    zorder_offset=0,
    ls_1d="-",
    alpha_1d=1.0,
    alpha=1.0,
    set_axlims=True,
    return_handle=False,
    show_ticks=True,
    lw_1d=0.75,
    **kwargs,
):
    """
    Create a triangle plot with contours.
    This can be called multiple times to add multiple data sets to one plot. For subsequent calls pass the axes
    instance and (you probably want to) set set_axlims=False.

    Parameters
    --------------
    pars_to_plot : list
        List of parameters to plot.
    bf_point : dict
        Best-fit point, used to centre the confidence ellipses.
    par_cov : array
        Parameter covariance matrix.
    pars_in_cov : list
        List of parameters in par_cov.
    sigma_plot_range : float
        Used to set axes limits.
    sigma_levels : int
        How far out to show contours.
    ax :
        if None makes new figure, if matplotlib.axes adds to passed axes.
    contour_colours: str/array
        Face colours of the contours.
    contour_edgecolour: str/array
        Edge colours of the contours.
    zorder_offset : int
        Offset to be applied to zorder of confidence ellipses. Handy to get right background/foreground ordering.
    ls_1d : str
        Linestyle for 1d panels.
    alpha_1d : float
        Alpha for the 1d lines.
    alpha : float
        Alpha to be applied to the confidence ellipses.
    set_axlims : bool
        Whether to set the ax lims or not. Likely only want this the first time this function is called on a given
        ax instance.
    return_handle : bool
        Whether to return a handle to create a legend with.
    show_ticks : bool
        Whether to show ticks (switch off to blind results).
    lw_1d: float
        Linewidth for 1d panels.
    kwargs:
        Other arguments to be passed through to matplotlib.patches.Ellipse.
    """

    # Create subplots
    if ax is None:
        fig, ax = plt.subplots(
            len(pars_to_plot),
            len(pars_to_plot),
            figsize=(2 * 3.464, 2 * 3.464),
            gridspec_kw={"hspace": 0.0, "wspace": 0.0},
        )

    # Plotting setup
    if contour_colours is None:
        contour_colours = sns.color_palette("Greys", sigma_levels + 1)[:-1][::-1]

    # Step through parameters
    for i, p1 in enumerate(pars_to_plot):
        for j, p2 in enumerate(pars_to_plot):
            # Only one side of triangle needed
            if i < j:
                ax[i, j].axis("off")
                continue

            # Convert between indices of pars to plot and in minimiser run
            i_par = pars_in_cov.index(p1)
            j_par = pars_in_cov.index(p2)

            if i == j:
                # Plot 1D Gaussian centred on bf point
                if set_axlims:
                    # Important scaling for 2 dimensions - otherwise contours are too tight.
                    # See information here https://arxiv.org/pdf/0906.4123.pdf
                    prob_amount = sp.stats.norm.cdf(sigma_levels) - sp.stats.norm.cdf(
                        -sigma_levels
                    )
                    scale_factor = np.sqrt(sp.stats.chi2.ppf(prob_amount, 2))

                    par_range = np.linspace(
                        -np.sqrt(np.diag(par_cov)[i_par])
                        * sigma_plot_range
                        * scale_factor
                        / 2.0,
                        np.sqrt(np.diag(par_cov)[i_par])
                        * sigma_plot_range
                        * scale_factor
                        / 2.0,
                        100,
                    )
                else:
                    par_range = np.linspace(
                        ax[i, j].get_xlim()[0], ax[i, j].get_xlim()[1], 100
                    )
                    par_range -= bf_point[p1]

                line_colour = contour_colours[0]
                if line_colour == "None":
                    line_colour = contour_edgecolour

                norm = sp.stats.norm.pdf(0, scale=np.sqrt(np.diag(par_cov)[i_par]))
                (handle,) = ax[i, j].plot(
                    par_range + bf_point[p1],
                    sp.stats.norm.pdf(par_range, scale=np.sqrt(np.diag(par_cov)[i_par]))
                    / norm,
                    color=line_colour,
                    alpha=alpha_1d,
                    ls=ls_1d,
                    lw=lw_1d,
                )

            else:
                # Plot sigma contours
                for n_std in range(sigma_levels):
                    add_confidence_ellipse(
                        ax[i, j],
                        par_cov,
                        j_par,
                        i_par,
                        bf_point[p2],
                        bf_point[p1],
                        n_std + 1,
                        facecolor=contour_colours[n_std],
                        zorder=-n_std + zorder_offset,
                        alpha=alpha,
                        edgecolor=contour_edgecolour,
                        **kwargs,
                    )

            # set axes limits
            if set_axlims:
                # Important scaling for 2 dimensions - otherwise contours are too tight.
                # See information here https://arxiv.org/pdf/0906.4123.pdf
                prob_amount = sp.stats.norm.cdf(sigma_levels) - sp.stats.norm.cdf(
                    -sigma_levels
                )
                scale_factor = np.sqrt(sp.stats.chi2.ppf(prob_amount, 2))

                ax[i, j].set_xlim(
                    (
                        bf_point[p2]
                        - np.sqrt(np.diag(par_cov)[j_par])
                        * sigma_plot_range
                        * scale_factor
                        / 2.0,
                        bf_point[p2]
                        + np.sqrt(np.diag(par_cov)[j_par])
                        * sigma_plot_range
                        * scale_factor
                        / 2.0,
                    )
                )

                if i != j:
                    ax[i, j].set_ylim(
                        (
                            bf_point[p1]
                            - np.sqrt(np.diag(par_cov)[i_par])
                            * sigma_plot_range
                            * scale_factor
                            / 2.0,
                            bf_point[p1]
                            + np.sqrt(np.diag(par_cov)[i_par])
                            * sigma_plot_range
                            * scale_factor
                            / 2.0,
                        )
                    )

            # Plotting stuff
            if i == len(pars_to_plot) - 1:
                par_lbl = p2
                if p2 in list(PAR_LABEL_DICT.keys()):
                    par_lbl = rf"${PAR_LABEL_DICT[p2]}$"
                ax[i, j].set_xlabel(par_lbl)
                if not show_ticks:
                    ax[i, j].set_xticklabels([])
            else:
                ax[i, j].set_xticklabels([])

            if j == 0 and i != j:
                par_lbl = p1
                if p1 in list(PAR_LABEL_DICT.keys()):
                    par_lbl = rf"${PAR_LABEL_DICT[p1]}$"
                ax[i, j].set_ylabel(par_lbl)
                ax[i, j].yaxis.set_label_coords(-0.45, 0.5)
                if not show_ticks:
                    ax[i, j].set_yticklabels([])
                    ax[i, j].yaxis.set_label_coords(-0.3, 0.5)
            else:
                ax[i, j].set_yticklabels([])

            ax[i, j].minorticks_on()
            ax[i, j].xaxis.set_tick_params(
                which="both", bottom=True, top=i != j, direction="in"
            )

    if return_handle:
        return ax, handle

    return ax


def add_min_trajectory(
    ax,
    pars_to_plot,
    eval_points,
    par_cov,
    pars_in_cov,
    bf_point=None,
    base_colour="blue",
    dark_colours=False,
    markeredgewidth=0.25,
):
    """
    Add the trajectory of a minimiser run to a triangle plot. Designed to work with plots created by
    triangle_plot_from_cov().

    Parameters
    --------------
    ax : matplotlib.axes
        The axes to add the ellipse to.
    pars_to_plot : list
        List of parameters to plot.
    eval_points : list
        List of parameter dictionaries giving the minimiser steps.
    par_cov : array
        Parameter covariance matrix. This is only needed to set the correct height of the points in the 1d panels.
    bf_point : dict
        This is only needed to set the correct height of the points in the 1d panels.
    base_colour : str
        Base colour for the series of points.
    dark_colours : bool
        Whether points should go from colour->black (default is colour->white).
    markeredgewidth : float
        Thickness of the outline of points
    """

    if bf_point is None:
        bf_point = eval_points[-1]

    N_points = len(eval_points)

    # Make colour palette
    if dark_colours:
        all_points_colours = sns.dark_palette(base_colour, N_points)
        all_points_colours[0] = "k"
    else:
        all_points_colours = sns.light_palette(base_colour, N_points)
        all_points_colours[0] = "white"

    all_points_colours = all_points_colours[::-1]  # start in colour then go dark/light

    for i, p1 in enumerate(pars_to_plot):
        for j, p2 in enumerate(pars_to_plot):
            i_par = pars_in_cov.index(p1)
            j_par = pars_in_cov.index(p2)

            if i == j:
                # Plot path of minimisers (only individual points)!
                for i_newton in range(N_points):
                    ax[i, j].plot(
                        eval_points[i_newton][p1],
                        np.sqrt(2 * np.pi * np.diag(par_cov)[i_par])
                        * sp.stats.norm.pdf(
                            eval_points[i_newton][p1] - bf_point[p1],
                            scale=np.sqrt(np.diag(par_cov)[i_par]),
                        ),
                        color=all_points_colours[i_newton],
                        marker="o",
                        markeredgecolor="k",
                        markeredgewidth=markeredgewidth,
                        ms=4,
                        zorder=99,
                    )

            if i > j:
                # Plot 2D path of minimiser!
                ax[i, j].plot(
                    [eval_points[i_newton][p2] for i_newton in range(N_points)],
                    [eval_points[i_newton][p1] for i_newton in range(N_points)],
                    color="k",
                    lw=0.5,
                    zorder=50,
                )

                # Highlight the individual points
                for i_newton in range(N_points):
                    ax[i, j].plot(
                        eval_points[i_newton][p2],
                        eval_points[i_newton][p1],
                        color=all_points_colours[i_newton],
                        marker="o",
                        markeredgecolor="k",
                        markeredgewidth=markeredgewidth,
                        ms=4,
                        zorder=99,
                    )


# --------------------------------------#
# PLOT MINIMISER CONVERGENCE
# --------------------------------------#


def plot_minimiser_convergence(
    eval_points, pars_to_plot=None, par_cov=None, par_cov_order=None, relative=False
):
    """
    Plots the steps of the minimizer to help assess convergence.

    Parameters
    -----------------
    eval_points : list
        List of dictionaries representing the evaluation points.
    pars_to_plot : list, optional
        List of parameters to plot. If None, all parameters in eval_points will be plotted.
    par_cov : ndarray, optional
        Covariance matrix of the parameters. If provided, error bars will be plotted.
    par_cov_order : list, optional
        Order of parameters in the covariance matrix. Required if par_cov is provided.
    relative : bool, optional
        If True, the parameter values will be plotted relative to the last value and in units of the covariance. Default is False.
    """

    if not par_cov is None and par_cov_order is None:
        raise Exception(f"candl: if covariance is to be used, specify parameter oder.")

    if pars_to_plot is None:
        pars_to_plot = list(eval_points[0].keys())

    fig, ax = plt.subplots(
        len(pars_to_plot),
        1,
        sharex=True,
        sharey=False,
        gridspec_kw={"hspace": 0},
        figsize=(2 * 3.464, len(pars_to_plot) / 4 * 3.464),
    )

    for i, par in enumerate(pars_to_plot):
        par_vals = np.array([p[par] for p in eval_points])

        offset = 0
        scale = 1
        if relative:
            offset = par_vals[-1]
            scale = np.sqrt(par_cov[par_cov_order.index(par), par_cov_order.index(par)])

        ax[i].plot((par_vals - offset) / scale, color="k")

        # If covariance is passed, but relative is false, plot the error bars
        if not relative and not par_cov is None:
            for i_sigma in np.arange(1, 3):
                ax[i].axhline(
                    par_vals[-1]
                    + i_sigma
                    * np.sqrt(
                        par_cov[par_cov_order.index(par), par_cov_order.index(par)]
                    ),
                    color=["0.5", "0.7"][i_sigma - 1],
                )
                ax[i].axhline(
                    par_vals[-1]
                    - i_sigma
                    * np.sqrt(
                        par_cov[par_cov_order.index(par), par_cov_order.index(par)]
                    ),
                    color=["0.5", "0.7"][i_sigma - 1],
                )

        # ylabel
        par_label = par
        if par in PAR_LABEL_DICT:
            par_label = rf"${PAR_LABEL_DICT[par]}$"
        if relative:
            par_label = rf"$\Delta$" + par_label + rf"$/\sigma$"
        ax[i].set_ylabel(par_label)

        ax[i].yaxis.set_label_coords(-0.11, 0.5)

    # x axis
    ax[-1].set_xlabel("Minimiser Step")
    ax[-1].set_xlim(0, len(eval_points) - 1)


# --------------------------------------#
# PLOT BAND POWERS (TT/TE/EE)
# --------------------------------------#


def plot_band_powers(like, show_legend=True, colour_by_spec_type=False):
    """
    Plots the band powers for a given primary CMB likelihood.

    Parameters
    --------------
    like : candl.like
        The likelihood.
    show_legend : bool (optional)
        Whether to show the legend. Default is True.
    colour_by_spec_type bool (optional)
        Whether to colour the band powers by spectrum type. Default is False.

    """

    fig = plt.gcf()
    fig.set_size_inches(2 * 3.464, 1 * 3.464)

    if colour_by_spec_type:
        all_spec_type_colours = sns.color_palette(
            n_colors=len(np.unique(like.spec_types))
        )
        spec_type_colours = {
            np.unique(like.spec_types)[i]: all_spec_type_colours[i]
            for i in range(len(all_spec_type_colours))
        }

    for i, spec in enumerate(like.spec_order):
        # Grab plotting data
        ells_to_plot = like.effective_ells[like.bins_start_ix[i] : like.bins_stop_ix[i]]
        bandpowers_to_plot = like.data_bandpowers[
            like.bins_start_ix[i] : like.bins_stop_ix[i]
        ]
        error_bars_to_plot = np.sqrt(
            np.diag(like.covariance)[like.bins_start_ix[i] : like.bins_stop_ix[i]]
        )

        # Plot
        if colour_by_spec_type:
            lbl = like.spec_types[i]
            if lbl in like.spec_types[:i]:
                lbl = None
            plt.errorbar(
                ells_to_plot,
                abs(bandpowers_to_plot),
                error_bars_to_plot,
                lw=0,
                ms=2,
                marker="o",
                label=lbl,
                elinewidth=1,
                color=spec_type_colours[like.spec_types[i]],
            )
        else:
            plt.errorbar(
                ells_to_plot,
                abs(bandpowers_to_plot),
                error_bars_to_plot,
                lw=0,
                ms=2,
                marker="o",
                label=spec,
                elinewidth=1,
            )

    plt.yscale("log")

    if show_legend:
        plt.legend()

    plt.xlabel("Angular Multipole")
    plt.ylabel(r"$D_\ell \, [ \mu K^2 ]$")
    plt.title(like.name)


# --------------------------------------#
# PLOT MCMC CHAIN STEPS
# --------------------------------------#


def plot_mcmc_chain_steps(
    gd_samples,
    pars_to_plot,
    bf_point=None,
    par_cov=None,
    pars_in_cov=None,
    show_logl=True,
):
    """
    Plot steps from an MCMC chain. Intended to be used in conjunction with getdist and Cobaya/CosmoMC.

    Parameters
    --------------
    gd_samples : getdist.mcsamples.MCSamples
        Samples instance containing all the parameter values.
    pars_to_plot : list
        List of strings of parameter names to plot.
    bf_point : dict
        Best-fit point to indicate.
    par_cov : array (float)
        Parameter covariance matrix at the best-fit point.
    pars_in_cov : list
        List of strings specifying the order of parameters in the covariance matrix.
    show_logl : bool
        Whether to show the log likelihood values or not.
    """

    fig, ax = plt.subplots(
        len(pars_to_plot) + int(show_logl),
        1,
        figsize=(2 * 3.464, 2 * 3.464),
        sharex=True,
        gridspec_kw={"hspace": 0},
    )

    steps = np.arange(len(gd_samples.loglikes))

    # Plot logl
    if show_logl:
        ax[0].plot(steps, gd_samples.loglikes, color=sns.color_palette()[0])

        # y axis label
        ax[0].set_ylabel(r"$\log{\mathcal{L}}$")
        ax[0].yaxis.set_label_coords(-0.11, 0.5)

    # Plot parameters
    for i, par_to_plot in enumerate(pars_to_plot):
        sample_vals = gd_samples.samples[:, gd_samples.index[par_to_plot]]
        ax[i + int(show_logl)].plot(steps, sample_vals, color=sns.color_palette()[0])

        # Plot bf val if available
        if not bf_point is None:
            bf_val = bf_point[par_to_plot]
            ax[i + int(show_logl)].axhline(bf_val, color="0.7", lw=0.75, zorder=99)

            # Also plot 1 sigma band from parameter covariance if passed
            if not par_cov is None:
                fisher_err = np.sqrt(np.diag(par_cov))[pars_in_cov.index(par_to_plot)]
                for i_sigma in [-1, 1]:
                    ax[i + int(show_logl)].axhline(
                        bf_val + i_sigma * fisher_err,
                        color="0.7",
                        ls="--",
                        lw=0.75,
                        zorder=99,
                    )

        # y labels
        if par_to_plot in PAR_LABEL_DICT:
            ax[i + int(show_logl)].set_ylabel(rf"${PAR_LABEL_DICT[par_to_plot]}$")
        else:
            ax[i + int(show_logl)].set_ylabel(par_to_plot)
        ax[i + int(show_logl)].yaxis.set_label_coords(-0.11, 0.5)

    plt.xlim((np.amin(steps), np.amax(steps)))

    plt.xlabel("MCMC Step")

    return fig, ax


# --------------------------------------#
# FG COMPONENT PLOT (TT/TE/EE)
# --------------------------------------#


def plot_foreground_components(like, fg_dict):
    """
    Plot foreground components. Intended to be used with output from get_foreground_contributions() of candl.Like.
    Warning: this function is hardcoded for full ell range TT/TE/EE spectra 6 frequency combinations.
    If you have other needs please create a copy and modify it.
    This means this function can also not deal with cropped likelihoods.

    Parameters
    --------------
    like : candl.Like
        Likelihood instance. Used to get effective ell centres, spectrum order and identifiers, etc.
    fg_dict : dict
        Dictionary of spectra where each value is a dictionary holding the unbinned fg conitribution for a given
        source, i.e. the output of candl.Like.get_foreground_contributions() in the format "dict by spec".
    """

    # Figure out all foreground sources and associate them with colours
    all_fg_names = []
    for spec in fg_dict:
        all_fg_names += list(fg_dict[spec].keys())
    all_fg_names = np.unique(all_fg_names)
    all_colours = sns.color_palette(n_colors=len(all_fg_names))
    fg_colours = {all_fg_names[i]: all_colours[i] for i in range(len(all_fg_names))}

    # ylims tuned for TT/TE/EE on log scale
    ylims = {"TT": [1e-1, 1e4], "TE": [1e-2, 1e3], "EE": [1e-3, 1e2]}

    # Generate array of subplots
    fig, ax = plt.subplots(
        6,
        3,
        figsize=(2 * 3.464, 2 * 3.464),
        sharex=True,
        gridspec_kw={"hspace": 0, "wspace": 1 / 3},
    )

    # Loop over all spectra
    for i_spec, spec in enumerate(like.spec_order):
        # Grab plot indices
        ix = int(np.floor(i_spec / 3))
        jx = i_spec - ix * 3

        # Grab ell bins centres
        ell_bins = like.effective_ells[
            like.bins_start_ix[i_spec] : like.bins_stop_ix[i_spec]
        ]

        # Show data band powers
        ax[ix, jx].plot(
            ell_bins,
            abs(
                like.data_bandpowers[
                    like.bins_start_ix[i_spec] : like.bins_stop_ix[i_spec]
                ]
            ),
            color="k",
            lw=1,
        )

        # Loop over foreground components
        for fg in fg_dict[spec]:
            ax[ix, jx].plot(
                like.ells, fg_dict[spec][fg], label=fg, color=fg_colours[fg]
            )

        # Plotting stuff
        ax[ix, jx].set_yscale("log")
        ax[ix, jx].set_ylim(ylims[like.spec_types[i_spec]])
        ax[ix, jx].set_xlim((300, 4000))

        # Add legend for top row of plots
        if ix == 0:
            handles, labels = ax[ix, jx].get_legend_handles_labels()
            labels = ["\n".join(wrap(lbl, 20)) for lbl in list(labels)]
            ax[ix, jx].legend(
                handles, labels, bbox_to_anchor=(0, 1.4, 0.8, 1), loc="lower center"
            )

        # bottom x label
        if ix == 5:
            ax[ix, jx].set_xlabel("Angular Multipole $\ell$")

        # y frequency labels for the first column
        if jx == 0:
            ax[ix, jx].set_ylabel(
                rf"${'x'.join(like.spec_freqs[i_spec])}$".replace("x", " \\times ")
            )

        # Spectrum types for the first row
        if ix == 0:
            ax[ix, jx].set_title(
                f"$|D^{{{like.spec_types[i_spec]}}}_\ell| \; [\mathrm{{\mu K^2}}]$"
            )

    # Make room for legends
    plt.subplots_adjust(top=0.8)

    return fig, ax



"""
Various tools that interact with the likelihood and its products.

Overview:
--------------

Fisher matrix calculation:

* :func:`get_fisher_matrix`

Simple minimiser:

* :func:`newton_raphson_minimiser`
* :func:`newton_raphson_minimiser_bdp`
* :func:`add_uniform_scatter`

Mock data generation:

* :func:`generate_mock_data`

Multi-frequency analysis:

* :func:`get_foreground_contributions`
* :func:`make_MV_combination`
* :func:`test_statistic_MV_consistency`
* :func:`make_frequency_conditional`
* :func:`conditional_prediction`
* :func:`test_statistic_conditional`
* :func:`make_difference_spectra`
* :func:`test_statistic_difference`

Bundling likelihood and theory code together:

* :func:`get_params_to_logl_func`
* :func:`get_params_to_chi_square_func`

Helpers to transform model spectra:

* :func:`pars_to_model_specs`
* :func:`pars_to_model_specs_partial_transformation`
* :func:`undo_transformations`
"""

# --------------------------------------#
# IMPORTS
# --------------------------------------#

from candl.lib import *
import candl
import candl.transformations.abstract_base
import candl.io

# --------------------------------------#
# FISHER MATRIX
# --------------------------------------#


def get_fisher_matrix(
    pars_to_theory_specs,
    like,
    pars,
    par_order=None,
    return_par_order=False,
    return_par_cov=True,
):
    """
    Calculate the Parameter covariance matrix based on the Fisher matrix using a differentiable theory code and the covariance of the data set.
    Applies priors on nuisance parameters.

    Parameters
    --------------
    pars_to_theory_specs : func
        Differentiable function moving dictionary of parameters to CMB spectra.
    like : candl.Like
        Likelihood
    pars : dict
        Dictionary of parameter values.
    par_order : list (optional)
        Order of parameters requested in returned matrix.
    return_par_order : bool (optional)
        return a list of parameters to indicate the order in the matrix.
    return_par_cov : bool (optional)
        Return the parameter covariance matrix, setting to False will return the Fisher matrix (i.e. not inverted).

    Returns
    --------------
    array (float) :
        Parameter covariance matrix.
    list (str) :
        List of parameter names.
    """

    # Define parameter order
    if par_order is None:
        par_order = list(pars.keys())

    # Get derivate function
    model_specs = lambda p: pars_to_model_specs(like, p, pars_to_theory_specs)[1]
    model_specs_deriv = jacfwd(
        model_specs
    )  # Forward autodiff should be better here (going from few cosmological parameters to many band power bins)

    # Calculate derivatives
    Dl_derivs_dict = model_specs_deriv(pars)
    Dl_derivs_arr = []
    for j, p in enumerate(par_order):
        Dl_derivs_arr.append(Dl_derivs_dict[p])
    Dl_derivs_arr = np.asarray(Dl_derivs_arr)

    # Calculate Fisher matrix
    fisher_mat = np.dot(
        Dl_derivs_arr, np.dot(np.linalg.inv(like.covariance), Dl_derivs_arr.T)
    )

    # Apply priors
    for prior in like.priors:
        prior_cov_inv = np.linalg.inv(prior.prior_covariance)
        for i, p1 in enumerate(prior.par_names):
            for j, p2 in enumerate(prior.par_names):
                try:
                    fisher_mat[
                        par_order.index(p1), par_order.index(p2)
                    ] += prior_cov_inv[i, j]
                except:
                    pass

    # Invert and return
    if return_par_cov:
        par_cov = np.linalg.inv(fisher_mat)

        if return_par_order:
            return par_cov, par_order
        return par_cov
    else:
        if return_par_order:
            return fisher_mat, par_order
        return fisher_mat


# --------------------------------------#
# MINIMISATION
# --------------------------------------#


def newton_raphson_minimiser(
    like_deriv,
    like_hess,
    starting_pars,
    pars_for_min,
    N_newton=10,
    step_size=0.2,
    show_progress=False,
):
    """
    Take Newton-Raphson steps towards the minimum.

    Warning
    --------------
    This is not a stress-tested minimiser, but rather a simple implementation of the algorithm. Use at your own risk.

    Parameters
    --------------
    like_deriv : func
        Derivative of the likelihood.
    like_hess : func
        Hessian of the likelihood.
    starting_pars : dict
        Start parameters.
    pars_for_min : list
        List of parameters to minimise.
    N_newton : int
        Number of steps to take.
    step_size : float
        Scaling for step size.
    show_progress : bool
        Show tqdm progress bar.

    Returns
    --------------
    list :
        List of all evaluation points
    list :
        List of parameter covariance matrix at all evaluation points.
    """

    # Progress bar
    if show_progress:
        pbar = tqdm(total=N_newton)

    # Prepare for Newton steps
    eval_points = [starting_pars]
    eval_par_cov = []
    has_failed = False
    for i_newton in range(N_newton + 1):
        # Calculate Hessian
        this_hess_dict = like_hess(eval_points[-1])

        # Turn into a nice array
        this_hess_arr = np.zeros((len(pars_for_min), len(pars_for_min)))
        for i, p1 in enumerate(pars_for_min):
            for j, p2 in enumerate(pars_for_min):
                this_hess_arr[i, j] = this_hess_dict[p1][p2]

        # Get the inverse
        this_par_cov = np.linalg.inv(-this_hess_arr)
        eval_par_cov.append(this_par_cov)

        # Exit here if desired number of steps have been reached
        if i_newton == N_newton:
            break

        # Juggle dict to arr
        par_dict = like_deriv(eval_points[i_newton])
        par_arr = np.zeros(len(pars_for_min))
        for j, p in enumerate(pars_for_min):
            par_arr[j] = par_dict[p]

        # Calculate Newton step
        step = jnp.dot(this_par_cov, par_arr)

        # Back into dict form for next step
        new_pars = deepcopy(eval_points[i_newton])
        for j, p in enumerate(pars_for_min):
            if step[j] is jnp.nan:  # stop if nan is encountered
                has_failed = True
                break
            new_pars[p] += step_size * step[j]

        # If step failed break out of loop
        if has_failed:
            break

        # Add any fixed parameter values
        for p in list(starting_pars.keys()):
            if not p in pars_for_min:
                new_pars[p] = starting_pars[p]

        eval_points.append(new_pars)

        if show_progress:
            pbar.update(1)

    if show_progress:
        pbar.close()

    # If minimiser has failed let the user know
    if has_failed:
        raise Exception(
            f"Minimiser failed (encountered nan in proposed step). Try a better initial point and smaller step size for better stability."
        )

    return eval_points, eval_par_cov


def newton_raphson_minimiser_bdp(
    like_deriv,
    like_hess,
    starting_pars,
    pars_for_min,
    bdp,
    N_newton=10,
    step_size=1,
    show_progress=False,
):
    """
    Same as newton_raphson_minimiser() but allows for the input of data band powers.
    This can be useful if dealing with the same likelihood but lots of mock data realisations to avoid the jitting
    penalty of new likelihoods for each realisation.

    Warning
    --------------
    This is not a stress-tested minimiser, but rather a simple implementation of the algorithm. Use at your own risk.
    """

    # Progress bar
    if show_progress:
        pbar = tqdm(total=N_newton)

    # Prepare for Newton steps
    eval_points = [starting_pars]
    eval_par_cov = []
    has_failed = False
    for i_newton in range(N_newton + 1):
        # Calculate Hessian
        this_hess_dict = like_hess(eval_points[-1], bdp)

        # Turn into a nice array
        this_hess_arr = jnp.zeros((len(pars_for_min), len(pars_for_min)))
        for i, p1 in enumerate(pars_for_min):
            for j, p2 in enumerate(pars_for_min):
                this_hess_arr = jax_optional_set_element(
                    this_hess_arr, (i, j), this_hess_dict[p1][p2]
                )
                # this_hess_arr[i, j] = this_hess_dict[p1][p2]

        # Get the inverse
        this_par_cov = jnp.linalg.inv(-this_hess_arr)
        eval_par_cov.append(this_par_cov)

        # Exit here if desired number of steps have been reached
        if i_newton == N_newton:
            break

        # Juggle dict to arr
        par_dict = like_deriv(eval_points[i_newton], bdp)
        par_arr = jnp.zeros(len(pars_for_min))
        for j, p in enumerate(pars_for_min):
            par_arr = jax_optional_set_element(par_arr, j, par_dict[p])
            # par_arr[j] = par_dict[p]

        # Calculate Newton step
        step = jnp.dot(this_par_cov, par_arr)

        # Back into dict form for next step
        new_pars = deepcopy(eval_points[i_newton])
        for j, p in enumerate(pars_for_min):
            if step[j] is jnp.nan:  # stop if nan is encountered
                has_failed = True
                break
            new_pars[p] += step_size * step[j]

        # If step failed break out of loop
        if has_failed:
            break

        # Add any fixed parameter values
        for p in list(starting_pars.keys()):
            if not p in pars_for_min:
                new_pars[p] = starting_pars[p]

        eval_points.append(new_pars)

        if show_progress:
            pbar.update(1)

    if show_progress:
        pbar.close()

    # If minimiser has failed let the user know
    if has_failed:
        raise Exception(
            f"Minimiser failed (encountered nan in proposed step). Try a better initial point and smaller step size for better stability."
        )

    return eval_points, eval_par_cov


def add_uniform_scatter(start_params, box_width, par_errs, seed=None):
    """
    Add uniform scatter to a point, helpful to offset starting points for minimisers.

    Parameters
    --------------
    start_params : dict
        Central parameter values.
    box_width : float
        Width of the sampled region in units of par_errs.
    par_errs : dict
        Parameter errors.
    seed : int
        RNG seed.

    Returns
    --------------
    dict :
        Offset parameters.
    dict :
        Parameter values delimiting the sampled region.
    """

    # Set RNG seed
    rng = np.random.default_rng(seed)

    # Add scatter and retain size of sampled region
    start_box = {}
    for par in par_errs:
        scatter_width = box_width * par_errs[par]
        par_scatter_range = [
            start_params[par] - scatter_width / 2,
            start_params[par] + scatter_width / 2,
        ]
        start_params[par] = rng.uniform(par_scatter_range[0], par_scatter_range[1])
        start_box[par] = par_scatter_range
    return start_params, start_box


# --------------------------------------#
# MOCK DATA GENERATION
# --------------------------------------#


def generate_mock_data(pars, pars_to_theory_specs, like, N_real, seed=None):
    """
    Generate mock band powers based on the covariance matrix.

    Parameters
    --------------
    pars : dict
        Fiducial parameter values.
    pars_to_theory_specs: func
        function moving dictionary of cosmological parameters to dictionary of CMB Dls
    like : Like
        Likelihood to take care of transformation and binning.
    N_real : int
        Number of realisations requested
    seed : int
        RNG seed.

    Returns
    --------------
    array :
        Binned model Dls (CMB and all transformation of the likelihood)
    list :
        List of mock band power realisations.
    """

    # Grab model spectrum
    unbinned_modified_theory_Dls, binned_modified_theory_Dls = pars_to_model_specs(
        like, pars, pars_to_theory_specs
    )

    # Add scatter
    rng = np.random.default_rng(seed)
    mock_bdp = []
    for i in range(N_real):
        this_scatter = jnp.dot(
            like.covariance_chol_dec,
            rng.standard_normal(np.shape(binned_modified_theory_Dls)[0]),
        )
        this_mock_bdp = binned_modified_theory_Dls + this_scatter
        mock_bdp.append(this_mock_bdp)

    return binned_modified_theory_Dls, mock_bdp


# --------------------------------------#
# FOREGROUND CONTRIBUTIONS
# --------------------------------------#


def get_foreground_contributions(
    like, params, return_type="arr", remove_zeros=True, binning="unbinned"
):
    """
    Return the foreground contributions.
    This method is intended to be accessed by the user for visualisation purposes to show the contributions of
    the different foreground components.

    Parameters
    --------------
    params : dict
        Nuisance parameter values.
    return_type : str
        One of ["arr", "dict", "dict by spec"]. Determines the output format.
    remove_zeros : bool
        Remove foreground components that are zero for all ells.
    binning : str
        One of ["unbinned", "binned", "both"]. Whether to return the foreground components for individual ells,
        binned, or both.

    Returns
    --------------
    Depending on return_type:
    "arr" : array, float
        A (N_ell_bins_theory*N_spectra_total or N_bins_total x N_foregrounds) array containing all the foreground
        contributions.
    "dict" : dict
        A dictionary with keys corresponding to the different foreground components and the values are
        (N_ell_bins_theory*N_spectra_total or N_bins_total) long vectors containing each contribution.
    "dict by spec" : dict
        A dictionary with keys corresponding to different spectra, each holding another dictionary with
        foreground names as keys and contributions as values.
    if binning is "both" then the output is wrapped in a dictionary containing "binned" and "unbinned" contributions
    at the lowest level.
    """

    # Check return_type argument
    if not return_type in ["arr", "dict", "dict by spec"]:
        raise Exception(
            f"Did not understand return_type '{return_type}'; must be one of (arr, dict, dict by spec). Returning as array"
        )
        return_type = "arr"

    # Check binning argument
    if not binning in ["unbinned", "binned", "both"]:
        raise Exception(
            f"Did not understand binning '{binning}'; must be one of (unbinned, binned, both). Returning unbinned"
        )
        binning = "unbinned"

    # Get foreground contributions
    foreground_contributions = []
    foreground_names = []
    for transformation in like.data_model:
        # Check if the transformation is a foreground - if so grab contribution and short name
        if isinstance(transformation, candl.transformations.abstract_base.Foreground):
            foreground_contributions.append(transformation.output(params))
            foreground_names.append(transformation.descriptor)

    foreground_contributions = jnp.asarray(foreground_contributions)

    # Return requested type
    if return_type == "dict":
        # Return as dictionary {label:contribution}
        if binning == "unbinned":
            return {
                foreground_names[i_fg]: foreground_contributions[i_fg, :]
                for i_fg in range(len(foreground_names))
            }
        elif binning == "binned":
            return {
                foreground_names[i_fg]: like.bin_model_specs(
                    foreground_contributions[i_fg, :]
                )
                for i_fg in range(len(foreground_names))
            }
        elif binning == "both":
            return {
                foreground_names[i_fg]: {
                    "unbinned": like.bin_model_specs(foreground_contributions[i_fg, :]),
                    "binned": like.bin_model_specs(foreground_contributions[i_fg, :]),
                }
                for i_fg in range(len(foreground_names))
            }
        return {
            foreground_names[i_fg]: foreground_contributions[i_fg, :]
            for i_fg in range(len(foreground_names))
        }

    elif return_type == "dict by spec":
        # Return as individual dictionaries for spectra {spec: {label:contribution}}
        dict_by_spec = {}
        for i_spec, spec in enumerate(like.spec_order):
            dict_by_spec[spec] = {}
            for i_fg, fg_label in enumerate(foreground_names):
                fg_arr = foreground_contributions[
                    i_fg,
                    like.N_ell_bins_theory
                    * i_spec : like.N_ell_bins_theory
                    * (i_spec + 1),
                ]
                if all(fg_arr == 0) and remove_zeros:
                    pass
                else:
                    if binning == "unbinned":
                        dict_by_spec[spec][fg_label] = fg_arr
                    else:
                        binned_fg_arr = like.bin_model_specs(
                            foreground_contributions[i_fg, :]
                        )[like.bins_start_ix[i_spec] : like.bins_stop_ix[i_spec]]
                        if binning == "binned":
                            dict_by_spec[spec][fg_label] = binned_fg_arr
                        elif binning == "both":
                            dict_by_spec[spec][fg_label] = {
                                "unbinned": fg_arr,
                                "binned": binned_fg_arr,
                            }
        return dict_by_spec

    # Return as array
    if binning == "unbinned":
        return foreground_contributions, foreground_names
    else:
        binned_fg_arr = []
        for i_fg in range(len(foreground_names)):
            binned_fg_arr.append(
                like.bin_model_specs(foreground_contributions[i_fg, :])
            )
        binned_fg_arr = jnp.asarray(binned_fg_arr)
        if binning == "binned":
            return binned_fg_arr, foreground_names
        elif binning == "both":
            return {
                "unbinned": foreground_contributions,
                "binned": binned_fg_arr,
            }, foreground_names


# --------------------------------------#
# MINIMUM-VARIANCE COMBINATION
# --------------------------------------#


def make_MV_combination(like, data_CMB_only, design_matrix, blinded=False):
    """
    Combine multifrequency band powers into MV estimate.
    See Appendix C4 here https://arxiv.org/pdf/1507.02704.pdf for details.

    Parameters
    --------------
    like : candl.Like
        The likelihood.
    data_CMB_only : array (float)
        The data bandpowers with all transformations undone.
    design_matrix : array (int)
        The design matrix.
    blinded : bool
        Whether operations are carried out in blinded or unblinded space (applies additional factor of the blinding function to the covariance).

    Returns
    --------------
    dict :
        "MV spec" : MV spectrum (array (float))
        "MV cov" : MV covariance (array (float))
        "MV ell" : effetive multipoles (array (float))
        "mix mat" : mixing matrix (array (float))
    """

    # Assemble MV band powers and covariance
    if blinded:
        cov_inv = np.linalg.inv(
            like.covariance * np.outer(like._blinding_function, like._blinding_function)
        )  # adjust covariance for blinding on the fly
    else:
        cov_inv = np.linalg.inv(like.covariance)

    weighting_fac = np.matmul(design_matrix.T, np.matmul(cov_inv, design_matrix))
    bdp_fac = np.matmul(design_matrix.T, np.matmul(cov_inv, data_CMB_only))

    MV_spec = np.matmul(np.linalg.inv(weighting_fac), bdp_fac)
    MV_cov = np.linalg.inv(
        np.matmul(design_matrix.T, np.matmul(cov_inv, design_matrix))
    )

    ell_eff_fac = np.matmul(design_matrix.T, np.matmul(cov_inv, like.effective_ells))
    MV_ell = np.matmul(np.linalg.inv(weighting_fac), ell_eff_fac)

    mixing_mat = np.linalg.inv(weighting_fac) @ design_matrix.T @ cov_inv

    return {
        "MV spec": MV_spec,
        "MV cov": MV_cov,
        "MV ell": MV_ell,
        "mix mat": mixing_mat,
    }


def test_statistic_MV_consistency(
    like, MV_dict, data_CMB_only, design_matrix, blinded=False
):
    """
    Test consistency by comparing multifrequency band powers to MV combination

    Parameters
    --------------
    like : candl.Like
        The likelihood.
    MV_dict : dict
        Dictionary containing information about MV combination (intended is output from make_MV_combination).
    data_CMB_only : array (float)
        The data bandpowers with all transformations undone.
    design_matrix : array (int)
        The design matrix.
    blinded : bool
        Whether operations are carried out in blinded or unblinded space (applies additional factor of the blinding function to the covariance).

    Returns
    --------------
    float:
        chisq
    float:
        PTE
    """

    delta = (design_matrix @ MV_dict["MV spec"]) - data_CMB_only

    if blinded:
        chol_fac = np.linalg.solve(
            np.linalg.cholesky(
                like.covariance
                * np.outer(like._blinding_function, like._blinding_function)
            ),
            delta,
        )  # adjust cov for blinding function
    else:
        chol_fac = np.linalg.solve(like.covariance_chol_dec, delta)
    chisq = np.dot(chol_fac.T, chol_fac)

    ndof = like.N_bins_total - len(MV_dict["MV spec"])
    PTE = 1 - sp.stats.chi2.cdf(chisq, ndof)

    return chisq, PTE


# --------------------------------------#
# FREQUENCY CONDITIONAL
# --------------------------------------#


def make_frequency_conditional(spec_str, like, best_fit_model_binned, blinded=False):
    """
    Generate frequency conditional prediction.
    See Section 3.6.3 here https://arxiv.org/pdf/1907.12875.pdf for details.

    Parameters
    --------------
    spec_str : str
        String identifying the spectrum. Intended format is (type) (freq1)x(freq2), e.g. "TT 150x150"
    like : candl.Like
        The likelihood.
    best_fit_model_binned : array (float)
        The binned best fit spectrum.
    blinded : bool
        Whether to operate in blinded space or not.

    Returns
    --------------
    dict :
        "cond spec" : Conditional spectrum (array (float))
        "cond cov" : Conditional covariance (array (float))
    """

    fld, freq_combo = spec_str.split(" ")

    # Make data vectors and covariance of all multifrequency spectra of this type
    fld_N_bins_list = [
        like.N_bins[i] for i in range(len(like.spec_order)) if like.spec_types[i] == fld
    ]
    fld_N_bins = sum(fld_N_bins_list)
    fld_start_ix, fld_stop_ix = candl.get_start_stop_ix(fld_N_bins_list)
    fld_spec_order = [
        like.spec_order[i]
        for i in range(len(like.spec_order))
        if like.spec_types[i] == fld
    ]

    fld_data_vec = np.zeros(fld_N_bins)
    fld_bf_vec = np.zeros(fld_N_bins)
    fld_cov = np.zeros((fld_N_bins, fld_N_bins))

    spec_msk = np.zeros(fld_N_bins)

    # Loop over all spectra and populate data structures
    for i, spec_1 in enumerate(fld_spec_order):
        full_vec_i = like.spec_order.index(spec_1)

        # data band powers
        if blinded:
            fld_data_vec[fld_start_ix[i] : fld_stop_ix[i]] = like.data_bandpowers[
                like.bins_start_ix[full_vec_i] : like.bins_stop_ix[full_vec_i]
            ]
        else:
            fld_data_vec[fld_start_ix[i] : fld_stop_ix[i]] = like._data_bandpowers[
                like.bins_start_ix[full_vec_i] : like.bins_stop_ix[full_vec_i]
            ]

        # model
        fld_bf_vec[fld_start_ix[i] : fld_stop_ix[i]] = best_fit_model_binned[
            like.bins_start_ix[full_vec_i] : like.bins_stop_ix[full_vec_i]
        ]

        if freq_combo == "x".join(like.spec_freqs[full_vec_i]):
            spec_msk[fld_start_ix[i] : fld_stop_ix[i]] = 1

        # covariance
        for j, spec_2 in enumerate(fld_spec_order):
            full_vec_j = like.spec_order.index(spec_2)

            this_cov_slice = like.covariance[
                like.bins_start_ix[full_vec_i] : like.bins_stop_ix[full_vec_i],
                like.bins_start_ix[full_vec_j] : like.bins_stop_ix[full_vec_j],
            ]

            if blinded:
                this_cov_slice *= np.outer(
                    like._blinding_function, like._blinding_function
                )[
                    like.bins_start_ix[full_vec_i] : like.bins_stop_ix[full_vec_i],
                    like.bins_start_ix[full_vec_j] : like.bins_stop_ix[full_vec_j],
                ]

            fld_cov[
                fld_start_ix[i] : fld_stop_ix[i], fld_start_ix[j] : fld_stop_ix[j]
            ] = this_cov_slice

    spec_msk = spec_msk == 1

    # Hand off to conditional function
    cond_spec, cond_cov = conditional_prediction(
        fld_data_vec, fld_bf_vec, fld_cov, spec_msk
    )

    return {"cond spec": cond_spec, "cond cov": cond_cov}


def conditional_prediction(x, x_bar, sigma, mask=None):
    """
    Linear algebra behind the conditional prediction.
    Code supplied by Karim Benabed.

    Parameters
    --------------
    x : array (float)
        The data vector.
    x_bar : array (float)
        The best-fit model.
    sigma : array (float)
        the covariance matrix
    mask : array (bool)
        Mask specifying which parts to excise in generating the conditional prediction.

    Returns
    --------------
    float:
        Conditional prediction
    float:
        Conditional covariance

    Notes
    --------------
    Original doc-string:
    This computes the conditional.
    x is the data vector, x_bar is the theoretical prediction.
    Both are 1D vectors and the ordering is whatever you want, provided that it's the same than the sigma covariance (a 2d array obviously).
    Mask is an optional 1D array of boolean, same size as x, to select a particular part of the data you want to ignore in the conditional implements eq. 353 of https://www.math.uwaterloo.ca/~hwolkowi/matrixcookbook.pdf.
    Returns, x_tilde and sigma_tilde.
    """

    if mask is None:
        mask = np.ones(len(x)) == 1
    xa = x_bar[mask]
    nmask = mask == False
    xb = x_bar[nmask]
    sigma_aa = sigma[mask, :][:, mask]
    sigma_bb = sigma[nmask, :][:, nmask]
    sigma_ab = sigma[mask, :][:, nmask]
    sigma_bb_inv = np.linalg.inv(sigma_bb)
    sigma_ab_DOT_sigma_bb_inv = np.dot(sigma_ab, sigma_bb_inv)
    x_tilde = xa + np.dot(sigma_ab_DOT_sigma_bb_inv, (x[nmask] - xb))
    sigma_tilde = sigma_aa - np.dot(sigma_ab_DOT_sigma_bb_inv, sigma_ab.T)

    return x_tilde, sigma_tilde


def test_statistic_conditional(spec_str, cond_dict, like, blinded=False):
    """
    Test conditional prediction against measured band powers.

    Parameters
    --------------
    spec_str : str
        String identifying the spectrum. Intended format is (type) (freq1)x(freq2), e.g. "TT 150x150"
    cond_dict : dict
        Dictionary containing information about the conditional prediction (intended is output from make_frequency_conditional).
    like : candl.Like
        The likelihood.
    blinded : bool
        Whether to operate on blinded band powers or not.

    Returns
    --------------
    float:
        chisq
    float:
        PTE
    """
    ix = like.spec_order.index(spec_str)
    if blinded:
        delta_cond = (
            cond_dict["cond spec"]
            - like.data_bandpowers[like.bins_start_ix[ix] : like.bins_stop_ix[ix]]
        )
    else:
        delta_cond = (
            cond_dict["cond spec"]
            - like._data_bandpowers[like.bins_start_ix[ix] : like.bins_stop_ix[ix]]
        )

    chisq = delta_cond.T @ np.linalg.inv(cond_dict["cond cov"]) @ delta_cond

    ndof = len(delta_cond)
    PTE = 1 - sp.stats.chi2.cdf(chisq, ndof)

    return chisq, PTE


# --------------------------------------#
# DIFFERENCE SPECTRA
# --------------------------------------#


def make_difference_spectra(spec_str_1, spec_str_2, data_CMB_only, like, blinded=False):
    """
    Generate frequency conditional prediction.
    See Section 3.6.3 here https://arxiv.org/pdf/1907.12875.pdf for details.

    Parameters
    --------------
    spec_str_1 : str
        String identifying the first spectrum. Intended format is (type) (freq1)x(freq2), e.g. "TT 150x150"
    spec_str_2 : str
        String identifying the second spectrum. Intended format is (type) (freq1)x(freq2), e.g. "TT 150x220"
    data_CMB_only : array (float)
        The data bandpowers with all transformations undone.
    like : candl.Like
        The likelihood.
    blinded : bool
        Whether to operate in blinded space.

    Returns
    --------------
    dict :
        "diff spec" : Difference spectrum (array (float))
        "diff cov" : Difference covariance (array (float))

    Warning
    --------------
    The code only works for equal length spectra at the moment.
    """

    # Check that valid difference was requested
    if spec_str_1[:2] != spec_str_1[:2] or spec_str_1 == spec_str_2:
        print(
            "Requested difference spectrum for different spectrum types or entered same spectrum twice!"
        )
        return

    # Grab indices
    i = like.spec_order.index(spec_str_1)
    j = like.spec_order.index(spec_str_2)

    # Grab spectra
    spec_1_fg_sub = data_CMB_only[like.bins_start_ix[i] : like.bins_stop_ix[i]]
    spec_2_fg_sub = data_CMB_only[like.bins_start_ix[j] : like.bins_stop_ix[j]]

    if len(spec_1_fg_sub) != len(spec_2_fg_sub):
        print("Requested spectra of different length, no design matrix supplied!")
        return

    diff = spec_1_fg_sub - spec_2_fg_sub

    # Grab cov
    cov_to_use = deepcopy(like.covariance)
    if blinded:
        cov_to_use *= np.outer(like._blinding_function, like._blinding_function)

    spec_1_cov = cov_to_use[
        like.bins_start_ix[i] : like.bins_stop_ix[i],
        like.bins_start_ix[i] : like.bins_stop_ix[i],
    ]
    spec_2_cov = cov_to_use[
        like.bins_start_ix[j] : like.bins_stop_ix[j],
        like.bins_start_ix[j] : like.bins_stop_ix[j],
    ]
    spec_12_cov = cov_to_use[
        like.bins_start_ix[i] : like.bins_stop_ix[i],
        like.bins_start_ix[j] : like.bins_stop_ix[j],
    ]
    spec_21_cov = cov_to_use[
        like.bins_start_ix[j] : like.bins_stop_ix[j],
        like.bins_start_ix[i] : like.bins_stop_ix[i],
    ]

    # Assemble covariance of the difference
    cov_block = np.block([[spec_1_cov, spec_12_cov], [spec_21_cov, spec_2_cov]])
    A_mat = np.block([[np.eye(len(spec_2_fg_sub)), -np.eye(len(spec_2_fg_sub))]])
    diff_cov = np.matmul(A_mat, np.matmul(cov_block, A_mat.T))

    return {"diff spec": diff, "diff cov": diff_cov}


def test_statistic_difference(diff_dict):
    """
    Test difference spectrum against zero.

    Parameters
    --------------
    diff_dict : dict
        Dictionary containing information about the difference spectrum (intended is output from make_difference_spectra).

    Returns
    --------------
    float:
        chisq
    float:
        PTE
    """

    chisq = (
        diff_dict["diff spec"].T
        @ np.linalg.inv(diff_dict["diff cov"])
        @ diff_dict["diff spec"]
    )
    ndof = len(diff_dict["diff spec"])
    PTE = 1 - sp.stats.chi2.cdf(chisq, ndof)

    return chisq, PTE


# --------------------------------------#
# BUNDLING LIKELIHOOD AND THEORY CODE TOGETHER
# --------------------------------------#


def get_params_to_logl_func(like, pars_to_theory_specs):
    """
    Thin wrapper bundling together likelihood and theory code to move straight from parameters to logl.

    Parameters
    --------------
    like: candl.Like
        Likelihood to be used.
    pars_to_theory_specs: func
        function moving dictionary of cosmological parameters to dictionary of CMB Dls

    Returns
    --------------
    func:
        Function that takes a dictionary of parameter values as input and returns the log likelihood
        (bundling theory code and likelihood together).

    """

    def like_with_theory(sampled_pars):
        # Calculate theory Dls
        Dls = pars_to_theory_specs(sampled_pars, like.ell_max, like.ell_min)

        # Pass to likelihood
        new_pars = sampled_pars
        new_pars["Dl"] = Dls
        logl = like.log_like(new_pars)

        return logl

    return like_with_theory


def get_params_to_chi_square_func(like, pars_to_theory_specs):
    """
    Thin wrapper bundling together likelihood and theory code to move straight from parameters to chi square.

    Parameters
    --------------
    like: candl.Like
        Likelihood to be used.
    pars_to_theory_specs: func
        function moving dictionary of cosmological parameters to dictionary of CMB Dls

    Returns
    --------------
    func:
        Function that takes a dictionary of parameter values as input and returns the chi square value
        (bundling theory code and likelihood together).

    """

    def chi_square_with_theory(sampled_pars):
        # Calculate theory Dls
        Dls = pars_to_theory_specs(sampled_pars, like.ell_max, like.ell_min)

        # Pass to likelihood
        new_pars = sampled_pars
        new_pars["Dl"] = Dls
        logl = like.chi_square(new_pars)

        return logl

    return chi_square_with_theory


# --------------------------------------#
# HELPERS TO TRANSFORM MODEL SPECTRA
# --------------------------------------#


def pars_to_model_specs(like, pars, pars_to_theory_specs):
    """
    Helper to move parameters to transformed model spectra and bin them.
    For lensing likelihoods only binned model spectra are returned.

    Parameters
    --------------
    like : candl.Like
        Likelihood to take care of transformation and binning.
    pars : dict
        Parameters for evaluation.
    pars_to_theory_specs: func
        function moving dictionary of cosmological parameters to dictionary of CMB Dls

    Returns
    --------------
    array:
        Best-fit model unbinned.
    array:
        Best-fit model binned.
    """

    # Calculate theory Dls
    Dls = pars_to_theory_specs(pars, like.ell_max, like.ell_min)

    # Pass to likelihood to apply transformations, bin
    new_pars = deepcopy(pars)
    new_pars["Dl"] = Dls

    model_unbinned = like.get_model_specs(new_pars)
    if isinstance(like, candl.LensLike):
        # For lensing likelihoods, this is already binned
        return model_unbinned, model_unbinned
    model_binned = like.bin_model_specs(model_unbinned)

    return model_unbinned, model_binned


def pars_to_model_specs_partial_transformation(
    like, pars, pars_to_theory_specs, i_tr_stop
):
    """
    Helper to move parameters to transformed model spectra and crop them.
    Only applies some of the transformations.

    Parameters
    --------------
    like : candl.Like
        Likelihood to take care of transformation and binning.
    pars : dict
        Parameters for evaluation.
    pars_to_theory_specs: func
        function moving dictionary of cosmological parameters to dictionary of CMB Dls
    i_tr_stop : int
        Index of the transformation at which to stop

    Returns
    --------------
    array:
        Model unbinned.
    array:
        Model binned.
    """

    # Calculate theory Dls
    Dls = pars_to_theory_specs(pars, like.ell_max, like.ell_min)
    new_pars = deepcopy(pars)
    new_pars["Dl"] = Dls

    # Unpack theory Dls into long vector
    modified_theory_Dls = jnp.block([new_pars["Dl"][st] for st in like.spec_types])

    # Apply nuisance modules
    for transformation in like.data_model[:i_tr_stop]:
        modified_theory_Dls = transformation.transform(modified_theory_Dls, new_pars)

    # bin if necessary
    if isinstance(like, candl.LensLike):
        # For lensing likelihoods, this is already binned
        return modified_theory_Dls, modified_theory_Dls
    best_fit_model_binned = like.bin_model_specs(modified_theory_Dls)

    return modified_theory_Dls, best_fit_model_binned


def undo_transformations(like, pars, pars_to_theory_specs):
    """
    Undo (best-fit) transformations from data vector.
    The order of the operations in the transformation matters - they do not commute.
    Hence, when undoing them, we must be careful to proceed in the correct order.
    The attribute 'operation_hint' of Transformation is used in order, if this attribute is not properly set this function will not return the correct results.
    In general, we need to undo calibration first, and then subtract the additive foregrounds.
    Special care needs to be taken for transformations that take Dls as their input too
    (e.g. super-sample lensing).

    Parameters
    --------------
    like : candl.Like
        Likelihood to take care of transformation and binning.
    pars : dict
        Parameters for evaluation.
    pars_to_theory_specs: func
        function moving dictionary of cosmological parameters to dictionary of CMB Dls

    Returns
    --------------
    array:
        Data band power with all transformations undone.
    """

    data_CMB_only_vec = deepcopy(like._data_bandpowers)

    # Undo transformations, one-by-one
    for i_tr, transformation in enumerate(like.data_model[::-1]):

        if transformation.operation_hint == "multiplicative":
            if isinstance(
                transformation, candl.transformations.abstract_base.Calibration
            ):
                # Catch calibration
                calibration_vec = transformation.transform(
                    np.ones(like.N_ell_bins_theory * like.N_spectra_total), pars
                )
                calibration_vec = np.repeat(
                    calibration_vec[:: like.N_ell_bins_theory], like.N_bins
                )

                data_CMB_only_vec /= calibration_vec
            else:
                # More general case
                unbinned_Dls, binned_Dls = pars_to_model_specs_partial_transformation(
                    like, pars, pars_to_theory_specs, len(like.data_model) - i_tr - 1
                )
                tr_vec_unbinned = transformation.transform(unbinned_Dls, pars)
                data_CMB_only_vec /= like.bin_model_specs(tr_vec_unbinned) / binned_Dls

        elif transformation.operation_hint == "additive":

            # Check if only sample parameters are required, or Dls too
            req_args_req = list(inspect.signature(transformation.output).parameters)
            if req_args_req == ["sample_params"]:
                tr_vec_unbinned = transformation.output(pars)
                data_CMB_only_vec -= like.bin_model_specs(tr_vec_unbinned)
            elif req_args_req == ["Dls", "sample_params"] or req_args_req == [
                "Dl",
                "sample_params",
            ]:
                # For the Dls we require the unbinned best-fit model Dls with all transformations up until now applied!
                unbinned_Dls, binned_Dls = pars_to_model_specs_partial_transformation(
                    like, pars, pars_to_theory_specs, len(like.data_model) - i_tr - 1
                )
                tr_vec_unbinned = transformation.output(unbinned_Dls, pars)
                data_CMB_only_vec -= like.bin_model_specs(tr_vec_unbinned)
            elif req_args_req == ["Dls"]:
                # Catches aberration
                unbinned_Dls, binned_Dls = pars_to_model_specs_partial_transformation(
                    like, pars, pars_to_theory_specs, len(like.data_model) - i_tr - 1
                )
                tr_vec_unbinned = transformation.output(unbinned_Dls)
                data_CMB_only_vec -= like.bin_model_specs(tr_vec_unbinned)
            else:
                print(
                    f"candl: not clear how to undo this transformation: {transformation}"
                )

        else:
            print(f"candl: not clear how to undo this transformation: {transformation}")

    return data_CMB_only_vec



"""
Main module containing the likelihood and prior classes.

Overview
-----------

Primary CMB likelihood class:

* :class:`Like`

Lensing likelihood class:

* :class:`LensLike`

Gaussian prior:

* :class:`GaussianPrior`

Misc:

* :func:`get_start_stop_ix`
* :func:`cholesky_decomposition`

"""

# --------------------------------------#
# IMPORTS
# --------------------------------------#

from candl.lib import *
import candl.io
import candl.transformations.abstract_base
import os

# --------------------------------------#
# GRAB DIRECTORY
# --------------------------------------#

candl_path = os.path.dirname(os.path.realpath(__file__))

# --------------------------------------#
# LIKELIHOOD
# --------------------------------------#


class Like:
    """
    Likelihood class for primary CMB power spectra (TT/TE/EE/BB) used to move from theory Dls and parameter values to log likelihood value.

    Attributes
    --------------
    N_bins : array, int
        The number of band power bins of each spectrum.
    N_bins_total : int
        The total number of band power bins.
    N_ell_bins_theory : int
        The number of ell bins (same for each spectrum).
    N_spectra_total : int
        The number of spectra.
    bins_start_ix : array, int
        The start indices for each spectrum in the long data vector.
    bins_stop_ix : array, int
        The stop indices for each spectrum in the long data vector.
    covariance : array, float
        The band power covariance matrix.
    covariance_chol_dec : array, float
        The cholesky decomposition of the band power covariance matrix.
    crop_mask : array, bool
        A mask to be applies to the original data vector for any subselection.
    data_bandpowers : array, float
        The data band powers, blinded if requested.
    _data_bandpowers : array, float
        The unblinded data band powers - careful!
    _blinding_function : array, float
        ratio of blinded to unblinded band powers - careful!
    data_model : list of candl.transformations.abstract_base.Transformation
        List of transformations to be applied to the theory spectra.
    data_set_dict : dict
        Dictionary read in from data set .yaml file containing all the information needed to initialise the likelihood.
    data_set_file : str
        The file path of the data set .yaml file.
    effective_ells : array, float
        Bin centres for each spectrum.
    effective_frequencies : dict
        Dictionary containing effective frequency information from the corresponding .yaml file.
    ells : array, float
        Angular multipole moments for a single spectrum.
    long_ells : array, float
         Angular multipole moments for all spectra.
    name : str
        Name of the likelihood.
    priors : list, GaussianPrior
        List of priors to be applied.
    required_nuisance_parameters : list, str
        List of nuisance parameters required for the likelihood.
    spec_freqs : list
        List of lists giving the two frequencies of each spectrum.
    spec_order : list
        List of strings specifying the order of spectra in the long data and model vectors.
    spec_types : list
        List of strings specifying the spectrum type (TT, TE, EE).
    unique_spec_types : list
        List of strings specifying the spectrum type (TT, TE, EE) without duplicate entries in no particular order.
    tiled_ells : array, float
        2d array of ell values for unbinned model spectra.
    window_functions : list of arrays (float)
        Band power window functions in order of spectra. One list entry for each spectrum with size (N_theory_bins, N_bins).
        Window functions start at ell=2.

    Methods
    --------------
    __init__ :
        initialises an instance of the class.
    log_like :
        Calculates the log likelihood given a theory CMB spectrum and nuisance parameter values.
    log_like_for_bdp :
        Calculates the log likelihood given a theory CMB spectrum, nuisance parameter values, and band powers.
    prior_logl :
        Calculates the log_like contribution of the priors.
    gaussian_logl :
        Gaussian log likelihood (helper for log_like)
    chi_square :
        Calculates the chi square given a theory CMB spectrum and nuisance parameter values.
    get_model_specs :
        Returns the theory CMB spectra adjusted by the nuisance parameter model.
    bin_model_specs :
        Performs the binning.
    init_data_model :
        Initialises the data model.
    init_priors :
        Initialises the priors.
    crop_for_data_selection :
        Adjusts the attributes of the likelihood for an arbitrary selection of the data.
    generate_crop_mask :
        Generate a boolean mask based on which data to use.
    interpret_crop_hint :
        Interpret a string hint to generate a crop mask.
    get_ell_helpers :
        Returns useful variables surrounding the ells.
    blind_bandpowers :
        Blinds bandpowers.

    """

    def __init__(self, data_set_file, **kwargs):
        """
        Initialise a new instance of the Like class.

        Parameters
        --------------
        data_set_file : str
            The file path of a .yaml file that contains all the necessary information to initialise the likelihood.
            Alternatively, this can be an index file that then points to the data set file. In this case, use the keyword "variant" to specify which variant of the data set to use.
        **kwargs : dict
            Any additional keyword arguments to overwrite the information in the .yaml file.

        Returns
        --------------
        Like
            A new instance of the base likelihood class with all data read in and the set-up completed.
        """

        # Load in the data set yaml file, checking if a variant is requested
        if not "variant" in kwargs:
            kwargs["variant"] = None
        self.data_set_dict, self.data_set_file = candl.io.load_info_yaml(
            data_set_file, kwargs["variant"]
        )

        # Overwrite any info in the data set yaml file by passed kwargs
        for key in kwargs:
            self.data_set_dict[key] = kwargs[key]

        # Specify data path if necessary
        if not "data_set_path" in self.data_set_dict:
            self.data_set_dict["data_set_path"] = (
                "/".join(self.data_set_file.split("/")[:-1]) + "/"
            )

        # Grab fluffy descriptor data
        self.name = candl.io.read_meta_info_from_yaml(self.data_set_dict)

        # Grab spectrum info
        (
            self.spec_order,
            self.spec_types,
            self.spec_freqs,
            self.N_spectra_total,
            self.N_bins,
        ) = candl.io.read_spectrum_info_from_yaml(self.data_set_dict)

        self.unique_spec_types = list(
            np.unique(self.spec_types)
        )  # order is destroyed in this process

        # Grab total number of spectra and bins
        self.N_spectra_total = len(self.spec_order)
        self.N_bins_total = sum(self.N_bins)

        # Get start and stop indices of each spectrum
        self.bins_start_ix, self.bins_stop_ix = get_start_stop_ix(self.N_bins)

        # Load band powers and covariance
        self._data_bandpowers = candl.io.read_file_from_yaml(
            self.data_set_dict, "band_power_file"
        )
        self.covariance = candl.io.read_file_from_yaml(
            self.data_set_dict, "covariance_file"
        )

        # Load beam correlation if present
        self.beam_correlation = None
        if "beam_correlation_file" in self.data_set_dict:
            self.beam_correlation = candl.io.read_file_from_yaml(
                self.data_set_dict, "beam_correlation_file"
            )

        # Set likelihood method to be used
        self.logl_function = self.gaussian_logl
        if "likelihood_form" in self.data_set_dict:
            if self.data_set_dict["likelihood_form"] == "gaussian_beam_detcov":
                if not self.beam_correlation is None:
                    self.logl_function = self.gaussian_logl_beam_and_detcov
        else:
            self.data_set_dict["likelihood_form"] = "gaussian"

        # Load effective frequencies
        self.effective_frequencies = candl.io.read_effective_frequencies_from_yaml(
            self.data_set_dict
        )

        # Read in band passes
        self.bandpasses = {}
        if "bandpasses" in self.data_set_dict:
            for freq in self.data_set_dict["bandpasses"]:
                bandpass_arr = candl.io.read_file_from_path(
                    self.data_set_dict["data_set_path"]
                    + self.data_set_dict["bandpasses"][freq]
                )
                this_bandpass = candl.transformations.abstract_base.BandPass(
                    bandpass_arr
                )
                self.bandpasses[str(freq)] = this_bandpass

        # Load in window functions and grab number of theory ell bins
        self.window_functions = candl.io.read_window_functions_from_yaml(
            self.data_set_dict, self.spec_order, self.N_bins
        )
        self.N_ell_bins_theory = int(jnp.shape(self.window_functions[0])[0])

        # Set ell range
        self.ell_min = 2  # everything starts at ell=2
        self.ell_max = self.N_ell_bins_theory + self.ell_min - 1

        # Preliminary ell helper (important to understand any ell based cropping about to happen)
        self.effective_ells = jnp.array([])
        for i, spec in enumerate(self.spec_order):
            self.effective_ells = jnp.concatenate(
                (
                    self.effective_ells,
                    jnp.dot(
                        jnp.arange(2, self.N_ell_bins_theory + 2),
                        self.window_functions[i],
                    ).flatten(),
                )
            )

        # Blind the band powers if necessary
        self.data_bandpowers = self._data_bandpowers
        if "blinding" in list(self.data_set_dict.keys()):
            if (
                type(self.data_set_dict["blinding"]) == bool
                and self.data_set_dict["blinding"]
            ) or type(self.data_set_dict["blinding"]) == int:
                self.data_bandpowers, self._blinding_function = self.blind_bandpowers(
                    (
                        None
                        if type(self.data_set_dict["blinding"]) == bool
                        else self.data_set_dict["blinding"]
                    ),
                )

        # Get a mask according to any subselection of the data
        self.crop_mask = self.generate_crop_mask()

        # Apply crop mask if needed
        # True == use bin
        # False == do not use bin
        if np.any(self.crop_mask == False):
            self.crop_for_data_selection()

        # Grab cholesky decomposition of the covariance
        # Check if Hartlap correction is requested
        if "hartlap_correction" in list(self.data_set_dict.keys()):
            self.covariance_chol_dec = cholesky_decomposition(
                self.covariance,
                N_sims=self.data_set_dict["hartlap_correction"]["N_sims"],
                N_bins=self.N_bins_total,
            )
        else:
            self.covariance_chol_dec = cholesky_decomposition(self.covariance)

        # Define ell range and grab some helpers
        (
            self.ells,
            self.tiled_ells,
            self.long_ells,
            self.effective_ells,
        ) = self.get_ell_helpers()

        # Expand any transformation blocks in the data model section of the .yaml file
        expanded_data_model_list = []
        for i, data_model_entry in enumerate(self.data_set_dict["data_model"]):
            if len(data_model_entry) == 1 and list(data_model_entry.keys()) == [
                "Block"
            ]:
                expanded_transformations = candl.io.expand_transformation_block(
                    self.data_set_dict, data_model_entry["Block"]
                )
                expanded_data_model_list.extend(expanded_transformations)
            else:
                expanded_data_model_list.append(data_model_entry)
        self.data_set_dict["data_model"] = expanded_data_model_list

        # Initialise the data model
        self.data_model = self.init_data_model()

        # Initialise any requested priors
        self.priors = self.init_priors()

        # Get a list of all the nuisance parameters the likelihood requires
        self.required_nuisance_parameters = []
        for transformation in self.data_model:  # go through all transformations
            for nuisance_par in transformation.param_names:
                self.required_nuisance_parameters.append(nuisance_par)
        self.required_nuisance_parameters = [
            str(p) for p in np.unique(self.required_nuisance_parameters)
        ]

        # Similarly, get a list of all parameters with priors
        self.required_prior_parameters = []
        for prior in self.priors:
            for prior_par in prior.par_names:
                self.required_prior_parameters.append(prior_par)
        self.required_prior_parameters = list(np.unique(self.required_prior_parameters))

        # Output info on initialisation
        candl.io.like_init_output(self)

    @partial(jit, static_argnums=(0,))
    def log_like(self, params):
        """
        Returns the negative log likelihood for a given set of theory Dls and nuisance parameter values.

        Parameters
        --------------
        params : dict
            Dictionary containing theory Dls and nuisance parameter values. Required entries: "Dl", the theory Dls starting from ell=2 and going to ell=2+N_ell_bins_theory. Nuisance parameters for any requested models.

        Returns
        --------------
        float
            Negative log likelihood.
        """

        # Get model spectra
        modified_theory_Dls = self.get_model_specs(params)

        # Bin long data vector
        binned_theory_Dls = self.bin_model_specs(modified_theory_Dls)

        # Calculate logl
        logl = self.logl_function(self._data_bandpowers, binned_theory_Dls)

        # Apply priors
        prior_logl = self.prior_logl(params)

        return -(logl + prior_logl)

    @partial(jit, static_argnums=(0,))
    def log_like_for_bdp(self, params, bdp):
        """
        Returns the negative log likelihood for a given set of theory Dls and nuisance parameter values.
        Shortcut to input band powers is useful to jit up the likelihood function with more flexibility,
        e.g. when calculating the derivative for different mock data sets.

        Parameters
        --------------
        params : dict
            Dictionary containing theory Dls and nuisance parameter values. Required entries: "Dl", the theory Dls starting from ell=2 and going to ell=2+N_ell_bins_theory. Nuisance parameters for any requested models.

        Returns
        --------------
        float
            Negative log likelihood.
        """

        # Get model spectra
        modified_theory_Dls = self.get_model_specs(params)

        # Bin long data vector
        binned_theory_Dls = self.bin_model_specs(modified_theory_Dls)

        # Calculate logl
        logl = self.logl_function(bdp, binned_theory_Dls)

        # Apply priors
        prior_logl = self.prior_logl(params)

        return -(logl + prior_logl)

    @partial(jit, static_argnums=(0,))
    def prior_logl(self, params):
        """
        Returns the positive log likelihood of the priors.

        Parameters
        --------------
        params : dict
            Dictionary containing nuisance parameter values.

        Returns
        --------------
        float
            Positive log likelihood.
        """
        p_logl = 0
        for prior in self.priors:
            p_logl += prior.log_like(params)
        return p_logl

    def gaussian_logl(self, data_bandpowers, binned_theory_Dls):
        """
        Calculate the positive log likelihood, i.e.:
        logl = 0.5 * (x-m) @ C^-1 @ (x-m)
        Uses the cholesky decomposition of the covariance for speed.

        Parameters
        --------------
        data_bandpowers : array, float
            Data band powers
        binned_theory_Dls : array, float
            Model spectra

        Returns
        --------------
        float
            Positive log likelihood.
        """

        # Calculate difference between model and data
        delta_bdp = data_bandpowers - binned_theory_Dls

        # Calculate logl
        chol_fac = jnp.linalg.solve(self.covariance_chol_dec, delta_bdp)
        chisq = jnp.dot(
            chol_fac.T, chol_fac
        )  # equivalent to the straightforward method, i.e. delta @ C^-1 @ delta
        logl = chisq / 2

        return logl

    def gaussian_logl_beam_and_detcov(self, data_bandpowers, binned_theory_Dls):
        """
        Calculate the positive log likelihood, including determinant of the covariance i.e.:
        logl = 0.5 * (x-m) @ C^-1 @ (x-m) + log | cov |
        Adds beam covariance and re-computes the Cholesky decomposition.

        Parameters
        --------------
        data_bandpowers : array, float
            Data band powers
        binned_theory_Dls : array, float
            Model spectra

        Returns
        --------------
        float
            Positive log likelihood.
        """

        # Calculate difference between model and data
        delta_bdp = data_bandpowers - binned_theory_Dls

        # Add beam contribution, recompute Cholesky decomposition
        full_covariance_chol_dec = jnp.linalg.cholesky(
            self.covariance
            + self.beam_correlation * jnp.outer(binned_theory_Dls, binned_theory_Dls)
        )

        # Apply Hartlap correction if requested
        if "hartlap_correction" in self.data_set_dict:
            full_covariance_chol_dec /= np.sqrt(
                (
                    self.data_set_dict["hartlap_correction"]["N_sims"]
                    - self.N_bins_total
                    - 2
                )
                / (self.data_set_dict["hartlap_correction"]["N_sims"] - 1)
            )

        # Calculate logl
        chol_fac = jnp.linalg.solve(full_covariance_chol_dec, delta_bdp)
        chisq = jnp.dot(
            chol_fac.T, chol_fac
        )  # equivalent to the straightforward method, i.e. delta @ C^-1 @ delta
        logl = chisq / 2
        logl += jnp.sum(
            jnp.log(jnp.diag(full_covariance_chol_dec))
        )  # covariance determinant term (only equal to half, but we are adding to the logl, not the chisq, so this is correct)

        return logl

    def chi_square(self, params):
        """
        Returns the chi squared for a given set of theory Dls and nuisance parameter values.
        Adds beam covariance contribution to covariance if present.

        Parameters
        --------------
        params : dict
            Dictionary containing theory Dls and nuisance parameter values. Required entries: "Dl", the theory Dls starting from ell=2 and going to ell=2+N_ell_bins_theory. Nuisance parameters for any requested models.

        Returns
        --------------
        float
            Chi square value
        """

        # Get model spectra
        modified_theory_Dls = self.get_model_specs(params)

        # Bin long data vector
        binned_theory_Dls = self.bin_model_specs(modified_theory_Dls)

        # Calculate difference between model and data
        delta_bdp = self._data_bandpowers - binned_theory_Dls

        # Add beam covariance contribution if necessary
        if not self.beam_correlation is None:
            full_covariance_chol_dec = jnp.linalg.cholesky(
                self.covariance
                + self.beam_correlation
                * jnp.outer(binned_theory_Dls, binned_theory_Dls)
            )
            chol_fac = jnp.linalg.solve(full_covariance_chol_dec, delta_bdp)
        else:
            # Calculate chisq
            chol_fac = jnp.linalg.solve(self.covariance_chol_dec, delta_bdp)

        chisq = jnp.dot(
            chol_fac.T, chol_fac
        )  # equivalent to the straight forward method, i.e. delta @ C^-1 @ delta

        return chisq

    @partial(jit, static_argnums=(0,))
    def get_model_specs(self, params):
        """
        Returns the theory spectra adjusted for the foreground/nuisance model.

        Parameters
        --------------
        params : dict
            Dictionary containing theory Dls and nuisance parameter values. Required entries: "Dl", the theory Dls starting from ell=2 and going to ell=2+N_ell_bins_theory. Nuisance parameters for any requested models.


        Returns
        --------------
        array (float)
            Theory Dls adjusted for the nuisance and foreground model. Covering ell=2-(2+N_ell_bins_theory).
        """

        # Unpack theory Dls into long vector
        modified_theory_Dls = jnp.block([params["Dl"][st] for st in self.spec_types])

        # Apply nuisance modules
        for transformation in self.data_model:
            modified_theory_Dls = transformation.transform(modified_theory_Dls, params)

        return modified_theory_Dls

    @partial(jit, static_argnums=(0,))
    def bin_model_specs(self, model_specs):
        """
        Bin model spectra.

        Parameters
        --------------
        model_specs : array, float
            The spectra to be binned.

        Returns
        --------------
        array (float)
            The binned spectra.
        """

        # Bin long data vector
        binned_theory_Dls = jnp.array([])
        for i, spec in enumerate(self.spec_order):
            binned_theory_Dls = jnp.concatenate(
                (
                    binned_theory_Dls,
                    jnp.dot(
                        model_specs[
                            i
                            * self.N_ell_bins_theory : (i + 1)
                            * self.N_ell_bins_theory
                        ],
                        self.window_functions[i],
                    ),
                )
            )

        return binned_theory_Dls

    def init_data_model(self):
        """
        Initialise the data model, i.e. the series of transformations to be applied to the theory spectra.
        Crucially, this function initialises all the transformation instances and passes them the keywords they need.
        It steps through the attributes needed by the initialisation of each class and supplies these from attributes
        of the likelihood as arguments or info from the relevant section of the .yaml file where the transformation.
        Additionally, it creates the freq_info list of effective frequencies and completes the path of the
        template_file.

        Returns
        --------------
        list of candl.transformations.abstract_base.Transformation
            The initialised transformation instances in the intended order ready to be applied to the theory spectra.
        """

        # Load in the data model
        data_model = []
        if not "data_model" in self.data_set_dict:
            return data_model

        for i_tr in range(len(self.data_set_dict["data_model"])):
            # Get class of the requested transformation and any passed arguments
            full_tr_name, tr_passed_args = candl.io.read_transformation_info_from_yaml(
                self.data_set_dict, i_tr
            )

            # Check whether transformation is deliberately ignored through backdoor
            ignore_transformation = False
            if "force_ignore_transformations" in self.data_set_dict:
                # loop over all transformations (identified by their descriptor) that should be ignored
                for ignore_tr_descriptor in (
                    self.data_set_dict["force_ignore_transformations"].split(", ")
                    if isinstance(
                        self.data_set_dict["force_ignore_transformations"], str
                    )
                    else self.data_set_dict["force_ignore_transformations"]
                ):
                    # This casts the different acceptable formats into a list of strings
                    if "descriptor" in tr_passed_args:
                        if tr_passed_args["descriptor"] == ignore_tr_descriptor:
                            ignore_transformation = True
            if ignore_transformation:
                continue

            # Break the information into module and class name
            module_name = ".".join(full_tr_name.split(".")[:-1])
            tr_name = full_tr_name.split(".")[-1]

            # Load the required module - either candl library or user's own
            try:
                try:
                    tr_module = importlib.import_module(
                        "candl.transformations." + module_name
                    )
                except:
                    tr_module = importlib.import_module(module_name)

                # Check what arguments are required to initialise the Transformation
                tr_signature = inspect.signature(tr_module.__dict__[tr_name]).parameters
                tr_all_args = list(tr_signature)
            except:
                raise Exception(f"Could not locate transformation: {full_tr_name}")

            # Compile all arguments required to initialise the transformation class
            tr_arg_dict = tr_passed_args  # use all passed args
            for arg in tr_all_args:
                # Add attributes from likelihood
                if arg in self.__dict__:
                    tr_arg_dict[arg] = self.__dict__[arg]

                # Add effective frequency information
                if arg == "freq_info":
                    freq_info = []
                    for nu_1, nu_2 in self.spec_freqs:
                        freq_info.append(
                            [
                                self.effective_frequencies[
                                    tr_passed_args["effective_frequencies"]
                                ][nu_1],
                                self.effective_frequencies[
                                    tr_passed_args["effective_frequencies"]
                                ][nu_2],
                            ]
                        )
                    tr_arg_dict["freq_info"] = freq_info
                    del tr_arg_dict["effective_frequencies"]

                # Add bandpass information if necessary
                if arg == "bandpass_info":
                    # Ensure all band passes have pre-calculated the thermodynamic conversion for this
                    # reference frequency
                    for band in self.bandpasses:
                        for (
                            passed_arg
                        ) in (
                            tr_arg_dict
                        ):  # some modules might have more than one reference frequency (e.g. correlation terms)
                            if passed_arg[:6] == "nu_ref":
                                if (
                                    not tr_arg_dict[passed_arg]
                                    in self.bandpasses[band].thermo_conv
                                ):
                                    self.bandpasses[
                                        band
                                    ].calculate_thermodynamic_conversion(
                                        tr_arg_dict[passed_arg]
                                    )

                    # Create list of band pass instances in correct order for spectra
                    bandpass_info = []
                    for nu_1, nu_2 in self.spec_freqs:
                        bandpass_info.append(
                            [self.bandpasses[nu_1], self.bandpasses[nu_2]]
                        )
                    tr_arg_dict["bandpass_info"] = bandpass_info

                # Read in any templates
                if "template_arr" in arg:
                    try:
                        tr_arg_dict[arg] = candl.io.read_file_from_path(
                            f"{candl_path}/{tr_arg_dict[arg.replace('template_arr', 'template_file')]}"
                        )
                    except:
                        try:
                            tr_arg_dict[arg] = candl.io.read_file_from_path(
                                f"{self.data_set_dict['data_set_path']}{tr_arg_dict[arg.replace('template_arr', 'template_file')]}"
                            )
                        except:
                            tr_arg_dict[arg] = candl.io.read_file_from_path(
                                f"{tr_arg_dict[arg.replace('template_arr', 'template_file')]}"
                            )
                    del tr_arg_dict[arg.replace("template_arr", "template_file")]

                # Link any already initialised transformations
                if arg[:26] == "link_transformation_module":
                    # Loop over already initialised modules and see if it's available

                    req_module_name = ".".join(tr_passed_args[arg].split(".")[:-1])
                    req_tr_name = tr_passed_args[arg].split(".")[-1]

                    try:
                        req_module = importlib.import_module(
                            "candl.transformations." + req_module_name
                        )
                    except:
                        req_module = importlib.import_module(req_module_name)

                    for transformation in data_model:
                        if isinstance(transformation, req_module.__dict__[req_tr_name]):
                            tr_arg_dict[arg] = transformation
                            break  # use the first available

                    if isinstance(tr_arg_dict[arg], str):
                        raise Exception(
                            f"Failed to initialise {tr_name} transformation."
                            f"Could not find a {tr_passed_args[arg]} instance to link to."
                        )

            # Initialise the transformation
            this_transformation = tr_module.__dict__[tr_name](**tr_arg_dict)
            data_model.append(this_transformation)

        return data_model

    def init_priors(self):
        """
        Initialise the priors. Similar to init_data_model, this function steps through the priors requested by the user
        and initialises GaussianPrior instances using the passed information.

        Returns
        --------------
        list of GaussianPrior
            The initialised priors ready to be evaluated.
        """

        priors = []
        if "priors" in self.data_set_dict:
            for prior_block in self.data_set_dict["priors"]:
                prior_args = prior_block

                if "prior_std" in prior_args:
                    prior_args["prior_covariance"] = float(prior_args["prior_std"]) ** 2
                    del prior_args["prior_std"]

                # Check if covariance file is passed rather than numerical value
                if isinstance(prior_args["prior_covariance"], str):
                    prior_args["prior_covariance"] = jnp.array(
                        np.loadtxt(
                            self.data_set_dict["data_set_path"]
                            + prior_args["prior_covariance"]
                        )
                    )
                else:
                    # avoid trouble parsing values with scientific 'e' notation
                    prior_args["prior_covariance"] = float(
                        prior_args["prior_covariance"]
                    )

                priors.append(GaussianPrior(**prior_args))

        return priors

    def crop_for_data_selection(self):
        """
        Adjust all relevant attributes of the likelihood for an arbitrary subselection of the data. This function
        manipulates not only all the information the likelihood stores on the number of bins, spectrum types etc.,
        but also crops the band powers, covariance, and window functions.
        """

        # Check if any spectra are completely deselected and can be removed entirely
        spec_deselected = [
            jnp.all(
                self.crop_mask[self.bins_start_ix[i_spec] : self.bins_stop_ix[i_spec]]
                == False
            )
            for i_spec in range(len(self.spec_order))
        ]

        # Update number of bins for each spectrum
        self.N_bins = [
            sum(self.crop_mask[self.bins_start_ix[i_spec] : self.bins_stop_ix[i_spec]])
            for i_spec in range(len(self.spec_order))
            if not spec_deselected[i_spec]
        ]

        # Delete deselected spectra from spectra order
        self.spec_order = [
            self.spec_order[i_spec]
            for i_spec in range(len(self.spec_order))
            if not spec_deselected[i_spec]
        ]

        # Infer types, frequencies, and total number of bins
        self.spec_types = [s[:2] for s in self.spec_order]
        self.N_spectra_total = len(self.spec_order)
        self.N_bins_total = jnp.sum(self.crop_mask)
        self.spec_freqs = [s.split(" ")[1].split("x") for s in self.spec_order]
        self.unique_spec_types = list(np.unique(self.spec_types))

        # Crop band powers, covariance matrix, window functions, and beam covariance (if present)
        self._data_bandpowers = jnp.delete(
            self._data_bandpowers, jnp.invert(self.crop_mask)
        )
        self.data_bandpowers = jnp.delete(
            self.data_bandpowers, jnp.invert(self.crop_mask)
        )
        self.covariance = jnp.delete(
            self.covariance, jnp.invert(self.crop_mask), axis=0
        )
        self.covariance = jnp.delete(
            self.covariance, jnp.invert(self.crop_mask), axis=1
        )

        # Crop window functions
        window_list = []
        for i, wdw in enumerate(self.window_functions):
            if not spec_deselected[i]:
                window_list.append(
                    jnp.delete(
                        wdw,
                        jnp.invert(self.crop_mask)[
                            self.bins_start_ix[i] : self.bins_stop_ix[i]
                        ],
                        axis=1,
                    )
                )
        self.window_functions = window_list

        # Crop beam correlation if present
        if not self.beam_correlation is None:
            self.beam_correlation = jnp.delete(
                self.beam_correlation, jnp.invert(self.crop_mask), axis=0
            )
            self.beam_correlation = jnp.delete(
                self.beam_correlation, jnp.invert(self.crop_mask), axis=1
            )

        # Update start and stop indices
        self.bins_start_ix, self.bins_stop_ix = get_start_stop_ix(self.N_bins)

    def generate_crop_mask(self):
        """
        Generate a crop mask. A crop mask is understood to be a boolean mask to be applied to the uncropped data vector
        with True entries for bins that should be used and False entries for bins that should be ignored.
        The method accessed the information on data selection passed by the user interprets it.

        If multiple string hints are passed then only bins selected by the most hints are used.

        See also: interpret_crop_hint

        Returns
        --------------
        array : bool
            The mask to be applied to the uncropped long data vector.
        """

        # Check if crop is necessary
        if not "data_selection" in self.data_set_dict:
            return np.ones(self.N_bins_total) == 1

        if (
            isinstance(self.data_set_dict["data_selection"], str)
            and self.data_set_dict["data_selection"].lower() == "none"
        ):
            return np.ones(self.N_bins_total) == 1

        if self.data_set_dict["data_selection"] is None:
            return np.ones(self.N_bins_total) == 1

        if isinstance(self.data_set_dict["data_selection"], str):
            if " " in self.data_set_dict["data_selection"]:
                # Generate crop mask based on some string hint
                crop_mask = self.interpret_crop_hint(
                    self.data_set_dict["data_selection"]
                )
            else:
                # Interpret as a relative path to a mask
                crop_mask = candl.io.read_file_from_yaml(
                    self.data_set_dict, "data_selection"
                )
        elif isinstance(self.data_set_dict["data_selection"], list) and isinstance(
            self.data_set_dict["data_selection"][0], str
        ):
            # Generate crop mask based on multiple string hints
            cumulative_crop_mask = np.zeros(self.N_bins_total)
            for this_string_hint in self.data_set_dict["data_selection"]:
                cumulative_crop_mask += self.interpret_crop_hint(this_string_hint)
            crop_mask = np.array(
                cumulative_crop_mask == np.amax(cumulative_crop_mask), dtype=bool
            )
        elif isinstance(self.data_set_dict["data_selection"], list) or isinstance(
            self.data_set_dict["data_selection"], np.ndarray
        ):
            # Can either be series of spectra to use or a boolean mask
            if isinstance(self.data_set_dict["data_selection"][0], str):
                crop_mask = np.array(
                    [
                        1 if spec in self.data_set_dict["data_selection"] else 0
                        for spec in self.spec_order
                    ]
                )
            elif (
                isinstance(self.data_set_dict["data_selection"][0], bool)
                or isinstance(self.data_set_dict["data_selection"][0], np.bool_)
            ) or (
                isinstance(self.data_set_dict["data_selection"][0], int)
                or isinstance(self.data_set_dict["data_selection"][0], np.int64)
            ):
                crop_mask = np.array(self.data_set_dict["data_selection"], dtype=int)
            else:
                # Cannot interpret the list passed
                crop_mask = np.ones(self.N_bins_total)
        else:
            crop_mask = np.ones(self.N_bins_total)
        return crop_mask == 1  # cast to bool array

    def interpret_crop_hint(self, crop_hint, verbose=False):
        """
        Interpret a string hint that indicates a crop mask.
        The intended format is: "(data) (action)", where "(data)" specifies which part of the data is selected and
        "(action)" declares what to do with this selection.
        Understood options for "(data)" are:
        * a specific spectrum as specified in the input .yaml file, e.g. "EE 90x90"
        * spectrum types, e.g. "TT"
        * frequencies and frequency combinations, e.g. "90" or "150x220"
        * ell range, e.g. "ell<650" or "ell>1500"
        Understood options for "(action)" are:
        * "remove" remove this part
        * "only" only keep this part, removing all the rest

        Returns
        --------------
        array : bool
            The mask to be applied to the uncropped long data vector corresponding to this data hint.
        """

        # Split hint into data and action part
        hint_list = crop_hint.split(" ")
        action_hint = hint_list[-1]
        data_hint = " ".join(hint_list[:-1])

        affected_specs = []
        base_msk = np.ones(self.N_bins_total) == 0

        if data_hint in self.spec_order:
            # Specific spectrum targeted
            for i_spec, spec in enumerate(self.spec_order):
                if data_hint == spec:
                    base_msk[self.bins_start_ix[i_spec] : self.bins_stop_ix[i_spec]] = (
                        True
                    )
                    affected_specs.append(self.spec_order[i_spec])

        elif data_hint in self.spec_types:
            # Probably concerns spectra
            for i_spec, st in enumerate(self.spec_types):
                if data_hint == st:
                    base_msk[self.bins_start_ix[i_spec] : self.bins_stop_ix[i_spec]] = (
                        True
                    )
                    affected_specs.append(self.spec_order[i_spec])

        elif data_hint.split("x")[0] in [f for freqs in self.spec_freqs for f in freqs]:
            # Probably concerns frequencies
            for i_spec, freq_pair in enumerate(self.spec_freqs):
                if set(data_hint.split("x")) == set(freq_pair):
                    base_msk[self.bins_start_ix[i_spec] : self.bins_stop_ix[i_spec]] = (
                        True
                    )
                    affected_specs.append(self.spec_order[i_spec])

        elif "ell" in data_hint:
            # ell cut
            ell_cut = float(data_hint.split("ell")[1][1:])
            compare_operation = data_hint.split("ell")[1][0]
            base_msk[eval(f"self.effective_ells{compare_operation}ell_cut")] = True

        else:
            raise Exception(
                f"Did not recognise data hint: {data_hint}\n Using all bins."
            )
            return np.ones(self.N_bins_total) == 1

        # Go through action hints and check whether we need to invert the selection
        if action_hint == "only":
            pass
        elif action_hint == "remove":
            base_msk = np.invert(base_msk)
        else:
            raise Exception(
                f"Did not recognise action: {action_hint}\n Using all bins."
            )
            return np.ones(self.N_bins_total) == 1

        # Feedback
        if len(affected_specs) > 0:
            specs_to_list = affected_specs
            if action_hint == "remove":
                specs_to_list = [
                    spec for spec in self.spec_order if not spec in affected_specs
                ]
            if verbose:
                print(f"Made a mask that keeps only: {specs_to_list}")

        return base_msk

    def get_ell_helpers(self):
        """
        Calculate useful variables concerning the ell range and binning.

        Returns
        --------------
        array : float
            array running over the angular multipoles for one spectrum (2->N_ell_bins_theory+2)
        array : float
            2d array giving the angular multipole range for each spectrum in rows
        array : float
            long array giving the angular multipole range for each spectrum (concatenated)
        array : float
            long array effective containing the effective bin centres (concatenated according to spectra_order).
        """

        ells = jnp.arange(2, self.N_ell_bins_theory + 2)
        tiled_ells = jnp.tile(ells, (self.N_spectra_total, 1))
        long_ells = tiled_ells.flatten()

        effective_ells = jnp.array([])
        for i, spec in enumerate(self.spec_order):
            effective_ells = jnp.concatenate(
                (
                    effective_ells,
                    jnp.dot(
                        jnp.arange(2, self.N_ell_bins_theory + 2),
                        self.window_functions[i],
                    ).flatten(),
                )
            )

        return ells, tiled_ells, long_ells, effective_ells

    def blind_bandpowers(self, seed=None):
        """
        Applies blinding function to bandpowers by multiplying by a random sinusoid function and a slope.
        It's important that the same spectrum types get hit by the same blinding function, i.e. e.g. the same for all TT spectra regardless of frequency.

        Parameters
        ----------
        seed : int, optional
            random seed to use

        Returns
        -------
        array
            blinded band powers
        array
            blinding function
        """

        rng = np.random.default_rng(seed)

        blind_func_by_spec = {}
        for spec_type in np.unique(self.spec_types):
            ells = None
            for i, spec in enumerate(self.spec_order):
                if spec[:2] == spec_type:
                    ells = self.effective_ells[
                        self.bins_start_ix[i] : self.bins_stop_ix[i]
                    ]
                    break

            blind_func = 1.0
            blind_func += rng.uniform(low=-0.9, high=0.9) * (
                (rng.integers(0, np.amax(ells)) - ells) / np.amax(ells)
            ) + rng.uniform(
                low=-0.1, high=0.1
            )  # tilt
            blind_func += rng.uniform(low=-0.5, high=0.5) * np.sin(
                ells / rng.integers(50, 100) + rng.integers(-100, 100)
            )  # acoustic peaks

            blind_func_by_spec[spec_type] = blind_func
        blind_func = np.block(
            [blind_func_by_spec[spec[:2]] for spec in self.spec_order]
        )

        return jnp.array(self._data_bandpowers * blind_func), jnp.array(blind_func)


# --------------------------------------#
# LENSING LIKELIHOOD
# --------------------------------------#


class LensLike:
    """
    Lensing likelihood class used to move from theory Dls and parameter values to the log likelihood value.

    Attributes
    -----------------
    N_bins : array, int
        The number of band power bins of each spectrum.
    N_bins_total : int
        The total number of band power bins.
    N_ell_bins_theory : int
        The number of ell bins (same for each spectrum).
    N_spectra_total : int
        The number of spectra.
    bins_start_ix : array, int
        The start indices for each spectrum in the long data vector.
    bins_stop_ix : array, int
        The stop indices for each spectrum in the long data vector.
    covariance : array, float
        The band power covariance matrix.
    covariance_chol_dec : array, float
        The cholesky decomposition of the band power covariance matrix.
    crop_mask : array, bool
        A mask to be applies to the original data vector for any subselection.
    data_bandpowers : array, float
        The data band powers, blinded if requested.
    _data_bandpowers : array, float
        The unblinded data band powers - careful!
    _blinding_function : array, float
        ratio of blinded to unblinded band powers - careful!
    data_model : list of candl.transformations.abstract_base.Transformation
        List of transformations to be applied to the theory spectra.
    data_set_dict : dict
        Dictionary read in from data set .yaml file containing all the information needed to initialise the likelihood.
    data_set_file : str
        The file path of the data set .yaml file.
    effective_ells : array, float
        Bin centres for each spectrum.
    ells : array, float
        Angular multipole moments for a single spectrum.
    lensing_fiducial_correction : array, float
        M * Cfid, loaded from files.
    long_ells : array, float
         Angular multipole moments for all spectra.
    name : str
        Name of the likelihood.
    priors : list, GaussianPrior
        List of priors to be applied.
    required_nuisance_parameters : list, str
        List of nuisance parameters required for the likelihood.
    spec_order : list
        List of strings specifying the order of spectra in the long data and model vectors.
    spec_types : list
        List of strings specifying the spectrum type (pp).
    unique_spec_types : list
        List of strings specifying the spectrum type (TT, TE, EE) without duplicate entries in no particular order.
    tiled_ells : array, float
        2d array of ell values for unbinned model spectra.
    window_functions : list of arrays (float)
        Band power window functions. Each entry in the N_specs long list is and array with size (N_theory_bins, N_bins).

    Methods
    ----------------
    __init__ :
        initialises an instance of the class.
    log_like :
        Calculates the log likelihood given a theory CMB spectrum and nuisance parameter values.
    log_like_for_bdp :
        Calculates the log likelihood given a theory CMB spectrum, nuisance parameter values, and band powers.
    prior_logl :
        Calculates the log_like contribution of the priors.
    gaussian_logl :
        Gaussian log likelihood (helper for log_like)
    chi_square :
        Calculates the chi square given a theory CMB spectrum and nuisance parameter values.
    get_model_specs :
        Returns the theory CMB spectra adjusted by the nuisance parameter model.
    bin_model_specs :
        Performs the binning.
    init_data_model :
        Initialises the data model.
    init_priors :
        Initialises the priors.
    crop_for_data_selection :
        Adjusts the attributes of the likelihood for an arbitrary selection of the data.
    generate_crop_mask :
        Generate a boolean mask based on which data to use.
    interpret_crop_hint :
        Interpret a string hint to generate a crop mask.
    get_ell_helpers :
        Returns useful variables surrounding the ells.
    blind_bandpowers :
        Blinds bandpowers.

    """

    def __init__(self, data_set_file, **kwargs):
        """
        Initialise a new instance of the LensLike class.

        Parameters
        --------------
        data_set_file : str
            The file path of a .yaml file that contains all the necessary information to initialise the likelihood.
            Alternatively, this can be an index file that then points to the data set file. In this case, use the keyword "variant" to specify which variant of the data set to use.
        kwargs : dict
            Any additional information to overwrite the information in the .yaml file.

        Returns
        --------------
        LensLike
            A new instance of the base lensing likelihood class with all data read in and the set-up completed.
        """

        # Load in the data set yaml file, checking if a variant is requested
        if not "variant" in kwargs:
            kwargs["variant"] = None
        self.data_set_dict, self.data_set_file = candl.io.load_info_yaml(
            data_set_file, kwargs["variant"]
        )

        # Overwrite any info in the data set yaml file by passed kwargs
        for key in kwargs:
            self.data_set_dict[key] = kwargs[key]

        # Specify data path if necessary
        if not "data_set_path" in self.data_set_dict:
            self.data_set_dict["data_set_path"] = (
                "/".join(self.data_set_file.split("/")[:-1]) + "/"
            )

        # Grab fluffy descriptor data
        self.name = candl.io.read_meta_info_from_yaml(self.data_set_dict)

        # Grab spectrum info
        (
            self.spec_order,
            self.spec_types,
            self.N_bins,
        ) = candl.io.read_spectrum_info_from_yaml(self.data_set_dict, lensing=True)

        self.unique_spec_types = list(
            np.unique(self.spec_types)
        )  # order is destroyed in this process

        # Grab total number of spectra and bins
        self.N_spectra_total = len(self.spec_order)
        self.N_bins_total = sum(self.N_bins)

        # Get start and stop indices of each spectrum
        self.bins_start_ix, self.bins_stop_ix = get_start_stop_ix(self.N_bins)

        # Load band powers and covariance
        self._data_bandpowers = candl.io.read_file_from_yaml(
            self.data_set_dict, "band_power_file"
        )
        self.covariance = candl.io.read_file_from_yaml(
            self.data_set_dict, "covariance_file"
        )

        # Load in window functions and grab number of theory ell bins
        self.window_functions = candl.io.read_window_functions_from_yaml(
            self.data_set_dict, self.spec_order, self.N_bins
        )
        self.N_ell_bins_theory = int(jnp.shape(self.window_functions[0])[0])

        # Set ell range
        self.ell_min = 2  # everything starts at ell=2
        self.ell_max = self.N_ell_bins_theory + self.ell_min - 1

        # Preliminary ell helper (important to understand any ell based cropping about to happen)
        self.effective_ells = jnp.array([])
        for i, spec in enumerate(self.spec_order):
            self.effective_ells = jnp.concatenate(
                (
                    self.effective_ells,
                    jnp.dot(
                        jnp.arange(2, self.N_ell_bins_theory + 2),
                        self.window_functions[i],
                    ).flatten(),
                )
            )

        # Blind the band powers if necessary
        self.data_bandpowers = self._data_bandpowers
        if "blinding" in list(self.data_set_dict.keys()):
            if (
                type(self.data_set_dict["blinding"]) == bool
                and self.data_set_dict["blinding"]
            ) or type(self.data_set_dict["blinding"]) == int:
                self.data_bandpowers, self._blinding_function = self.blind_bandpowers(
                    (
                        None
                        if type(self.data_set_dict["blinding"]) == bool
                        else self.data_set_dict["blinding"]
                    ),
                )

        # Get a mask according to any subselection of the data
        self.crop_mask = self.generate_crop_mask()

        # Apply crop mask if needed
        # True == use bin
        # False == do not use bin
        if np.any(self.crop_mask == False):
            self.crop_for_data_selection()

        # Grab cholesky decomposition of the covariance
        if "hartlap_correction" in list(self.data_set_dict.keys()):
            self.covariance_chol_dec = cholesky_decomposition(
                self.covariance,
                N_sims=self.data_set_dict["hartlap_correction"]["N_sims"],
                N_bins=self.N_bins_total,
            )
        else:
            self.covariance_chol_dec = cholesky_decomposition(self.covariance)

        # Define ell range and grab some helpers
        (
            self.ells,
            self.tiled_ells,
            self.long_ells,
            self.effective_ells,
        ) = self.get_ell_helpers()

        # Initialise the data model
        self.data_model = self.init_data_model()

        # Initialise any requested priors
        self.priors = self.init_priors()

        # Get a list of all the nuisance parameters the likelihood requires
        self.required_nuisance_parameters = []
        for transformation in self.data_model:  # go through all transformations
            for nuisance_par in transformation.param_names:
                self.required_nuisance_parameters.append(nuisance_par)
        self.required_nuisance_parameters = [
            str(p) for p in np.unique(self.required_nuisance_parameters)
        ]

        # Similarly, get a list of all parameters with priors
        self.required_prior_parameters = []
        for prior in self.priors:
            for prior_par in prior.par_names:
                self.required_prior_parameters.append(prior_par)
        self.required_prior_parameters = list(np.unique(self.required_prior_parameters))

        # Assert Likelihood form
        self.data_set_dict["likelihood_form"] = "gaussian"

        # Output info on initialisation
        candl.io.like_init_output(self)

    @partial(jit, static_argnums=(0,))
    def log_like(self, params):
        """
        Returns the negative log likelihood for a given set of theory Dls and nuisance parameter values.

        Parameters
        --------------
        params : dict
            Dictionary containing theory Dls and nuisance parameter values. Required entries: "Dl", the theory Dls starting from ell=2 and going to ell=2+N_ell_bins_theory. Nuisance parameters for any requested models.


        Returns
        --------------
        float
            Negative log likelihood.
        """

        # Get model spectra
        binned_theory_Dls = self.get_model_specs(params)

        # Calculate difference between model and data
        delta_bdp = self._data_bandpowers - binned_theory_Dls

        # Calculate logl
        logl = self.gaussian_logl(delta_bdp)

        # Apply priors
        prior_logl = self.prior_logl(params)

        return -(logl + prior_logl)

    @partial(jit, static_argnums=(0,))
    def log_like_for_bdp(self, params, bdp):
        """
        Returns the negative log likelihood for a given set of theory Dls and nuisance parameter values.
        Shortcut to input band powers is useful to jit up the likelihood function with more flexibility,
        e.g. when calculating the derivative for different mock data sets.

        Parameters
        --------------
        params : dict
            Dictionary containing theory Dls and nuisance parameter values. Required entries: "Dl", the theory Dls starting from ell=2 and going to ell=2+N_ell_bins_theory. Nuisance parameters for any requested models.


        Returns
        --------------
        float
            Negative log likelihood.
        """

        # Get model spectra
        binned_theory_Dls = self.get_model_specs(params)

        # Calculate difference between model and data
        delta_bdp = bdp - binned_theory_Dls

        # Calculate logl
        logl = self.gaussian_logl(delta_bdp)

        # Apply priors
        prior_logl = self.prior_logl(params)

        return -(logl + prior_logl)

    @partial(jit, static_argnums=(0,))
    def prior_logl(self, params):
        """
        Returns the positive log likelihood of the priors.

        Parameters
        --------------
        params : dict
            Dictionary containing nuisance parameter values.

        Returns
        --------------
        float
            Positive log likelihood.
        """
        p_logl = 0
        for prior in self.priors:
            p_logl += prior.log_like(params)
        return p_logl

    def gaussian_logl(self, delta_bdp):
        """
        Calculate the positive log likelihood, i.e.:
        logl = 0.5 * (x-m) @ C^-1 @ (x-m)
        Uses the cholesky decomposition of the covariance for speed.

        Parameters
        --------------
        delta_bdp : array, float
            Difference between data and model band powers

        Returns
        --------------
        float
            Positive log likelihood.
        """

        # Calculate logl
        chol_fac = jnp.linalg.solve(self.covariance_chol_dec, delta_bdp)
        chisq = jnp.dot(
            chol_fac.T, chol_fac
        )  # equivalent to the straightforward method, i.e. delta @ C^-1 @ delta
        logl = chisq / 2

        return logl

    def chi_square(self, params):
        """
        Returns the chi squared for a given set of theory Dls and nuisance parameter values.
        Intentionally not reusing the gaussian_logl function as we might add a beam covariance contribution
        at a later date.

        Parameters
        --------------
        params : dict
            Dictionary containing theory Dls and nuisance parameter values. Required entries: "Dl", the theory Dls starting from ell=2 and going to ell=2+N_ell_bins_theory. Nuisance parameters for any requested models.


        Returns
        --------------
        float
            Chi square value
        """

        # Get model spectra
        binned_theory_Dls = self.get_model_specs(params)

        # Calculate difference between model and data
        delta_bdp = self._data_bandpowers - binned_theory_Dls

        # Calculate chisq
        chol_fac = jnp.linalg.solve(self.covariance_chol_dec, delta_bdp)
        chisq = jnp.dot(
            chol_fac.T, chol_fac
        )  # equivalent to the straight forward method, i.e. delta @ C^-1 @ delta

        return chisq

    @partial(jit, static_argnums=(0,))
    def get_model_specs(self, params):
        """
        Returns the theory spectra adjusted for the foreground/nuisance model.

        Parameters
        --------------
        params : dict
            Dictionary containing theory Dls and nuisance parameter values. Required entries: "Dl", the theory Dls starting from ell=2 and going to ell=2+N_ell_bins_theory. Nuisance parameters for any requested models.


        Returns
        --------------
        array (float)
            Theory Dls adjusted for the nuisance and foreground model. Covering ell=2-(2+N_ell_bins_theory).
        """

        # Unpack theory Dls into long vector
        theory_Dls = jnp.block([params["Dl"][st] for st in self.spec_types])

        # Lensing foreground is already binned
        # Bin long data vector
        modified_theory_Dls = self.bin_model_specs(theory_Dls)

        # Apply nuisance modules
        for transformation in self.data_model:
            modified_theory_Dls = transformation.transform(modified_theory_Dls, params)

        return modified_theory_Dls

    @partial(jit, static_argnums=(0,))
    def bin_model_specs(self, model_specs):
        """
        Bin model spectra.

        Parameters
        --------------
        model_specs : array, float
            The spectra to be binned.

        Returns
        --------------
        array (float)
            The binned spectra.
        """

        # Bin long data vector
        binned_theory_Dls = jnp.array([])
        for i, spec in enumerate(self.spec_order):
            binned_theory_Dls = jnp.concatenate(
                (
                    binned_theory_Dls,
                    jnp.dot(
                        model_specs[
                            i
                            * self.N_ell_bins_theory : (i + 1)
                            * self.N_ell_bins_theory
                        ],
                        self.window_functions[i],
                    ),
                )
            )

        return binned_theory_Dls

    def init_data_model(self):
        """
        Initialise the data model, i.e. the series of transformations to be applied to the theory spectra.
        Crucially, this function initialises all the transformation instances and passes them the keywords they need.
        It steps through the attributes needed by the initialisation of each class and supplies these from attributes
        of the likelihood as arguments or info from the relevant section of the .yaml file where the transformation.
        Additionally, it creates the freq_info list of effective frequencies and completes the path of the
        template_file.

        Returns
        --------------
        list of candl.transformations.abstract_base.Transformation
            The initialised transformation instances in the intended order ready to be applied to the theory spectra.
        """

        # Load in the data model
        data_model = []
        if not "data_model" in self.data_set_dict:
            return data_model

        for i_tr in range(len(self.data_set_dict["data_model"])):
            # Get class of the requested transformation and any passed arguments
            full_tr_name, tr_passed_args = candl.io.read_transformation_info_from_yaml(
                self.data_set_dict, i_tr
            )

            # Check whether transformation is deliberately ignored through backdoor
            ignore_transformation = False
            if "force_ignore_transformations" in self.data_set_dict:
                # loop over all transformations (identified by their descriptor) that should be ignored
                for ignore_tr_descriptor in (
                    self.data_set_dict["force_ignore_transformations"].split(", ")
                    if isinstance(
                        self.data_set_dict["force_ignore_transformations"], str
                    )
                    else self.data_set_dict["force_ignore_transformations"]
                ):
                    # This casts the different acceptable formats into a list of strings
                    if "descriptor" in tr_passed_args:
                        if tr_passed_args["descriptor"] == ignore_tr_descriptor:
                            ignore_transformation = True
            if ignore_transformation:
                continue

            # Break the information into module and class name
            module_name = ".".join(full_tr_name.split(".")[:-1])
            tr_name = full_tr_name.split(".")[-1]

            # Load the required module - either candl library or user's own
            try:
                try:
                    tr_module = importlib.import_module(
                        "candl.transformations." + module_name
                    )
                except:
                    tr_module = importlib.import_module(module_name)

                # Check what arguments are required to initialise the Transformation
                tr_signature = inspect.signature(tr_module.__dict__[tr_name]).parameters
                tr_all_args = list(tr_signature)
            except:
                raise Exception(f"Could not locate transformation: {full_tr_name}")

            # Compile all arguments required to initialise the transformation class
            tr_arg_dict = tr_passed_args  # use all passed args
            for arg in tr_all_args:
                # Add attributes from likelihood
                if arg in self.__dict__:
                    tr_arg_dict[arg] = self.__dict__[arg]

                # Read in any templates
                if "template_arr" in arg:
                    try:
                        tr_arg_dict[arg] = candl.io.read_file_from_path(
                            f"{candl_path}/{tr_arg_dict[arg.replace('template_arr', 'template_file')]}"
                        )
                    except:
                        try:
                            tr_arg_dict[arg] = candl.io.read_file_from_path(
                                f"{self.data_set_dict['data_set_path']}{tr_arg_dict[arg.replace('template_arr', 'template_file')]}"
                            )
                        except:
                            tr_arg_dict[arg] = candl.io.read_file_from_path(
                                f"{tr_arg_dict[arg.replace('template_arr', 'template_file')]}"
                            )
                    del tr_arg_dict[arg.replace("template_arr", "template_file")]

                # Read in M matrices and fiducial correction, if needed
                if arg == "M_matrices":

                    # Start with M matrices, looping over requested corrections
                    M_matrices = dict()
                    for s in tr_arg_dict["Mmodes"]:
                        if s in ["TT", "TE", "EE", "BB", "pp", "kk"]:
                            M_matrices[s] = candl.io.read_lensing_M_matrices_from_yaml(
                                self.data_set_dict["data_set_path"]
                                + tr_arg_dict["M_matrices_folder"],
                                Mtype=s,
                            )
                    tr_arg_dict["M_matrices"] = M_matrices

                    # Now read in fiducial correction
                    fiducial_correction = candl.io.read_file_from_path(
                        self.data_set_dict["data_set_path"]
                        + tr_arg_dict["fiducial_correction_file"]
                    )

                    # If a matrix was passed, use corrections only for the requested spectra
                    if np.ndim(fiducial_correction) == 2:
                        spec_ix = {
                            "TT": 0,
                            "TE": 1,
                            "EE": 2,
                            "BB": 3,
                            "pp": 4,
                            "kk": 4,
                        }  # order expected in M_matrices and in fiducial corrections file
                        sum_msk = np.zeros(5, dtype=bool)
                        for s in list(spec_ix.keys()):
                            if s in tr_arg_dict["Mmodes"]:
                                sum_msk[spec_ix[s]] = True
                        fiducial_correction = jnp.sum(
                            fiducial_correction, axis=1, where=sum_msk
                        )

                    tr_arg_dict["fiducial_correction"] = fiducial_correction

                    # Clean up arguments
                    del tr_arg_dict["M_matrices_folder"]
                    del tr_arg_dict["Mmodes"]
                    del tr_arg_dict["fiducial_correction_file"]

            # Initialise the transformation
            this_transformation = tr_module.__dict__[tr_name](**tr_arg_dict)
            data_model.append(this_transformation)

        return data_model

    def init_priors(self):
        """
        Initialise the priors. Similar to init_data_model, this function steps through the priors requested by the user
        and initialises GaussianPrior instances using the passed information.

        Returns
        --------------
        list of GaussianPrior
            The initialised priors ready to be evaluated.
        """

        priors = []
        if "priors" in self.data_set_dict:
            for prior_block in self.data_set_dict["priors"]:
                prior_args = prior_block
                if "prior_std" in prior_args:
                    prior_args["prior_covariance"] = float(prior_args["prior_std"]) ** 2
                    del prior_args["prior_std"]

                # Check if covariance file is passed rather than numerical value
                if isinstance(prior_args["prior_covariance"], str):
                    prior_args["prior_covariance"] = jnp.array(
                        np.loadtxt(
                            self.data_set_dict["data_set_path"]
                            + prior_args["prior_covariance"]
                        )
                    )
                else:
                    # avoid trouble parsing values with scientific 'e' notation
                    prior_args["prior_covariance"] = float(
                        prior_args["prior_covariance"]
                    )

                priors.append(GaussianPrior(**prior_args))

        return priors

    def crop_for_data_selection(self):
        """
        Adjust all relevant attributes of the likelihood for an arbitrary subselection of the data. This function
        manipulates not only all the information the likelihood stores on the number of bins, spectrum types etc.,
        but also crops the band powers, covariance, and window functions.
        """

        # Check if any spectra are completely deselected and can be removed entirely
        spec_deselected = [
            jnp.all(
                self.crop_mask[self.bins_start_ix[i_spec] : self.bins_stop_ix[i_spec]]
                == False
            )
            for i_spec in range(len(self.spec_order))
        ]

        # Update number of bins for each spectrum
        self.N_bins = [
            sum(self.crop_mask[self.bins_start_ix[i_spec] : self.bins_stop_ix[i_spec]])
            for i_spec in range(len(self.spec_order))
            if not spec_deselected[i_spec]
        ]

        # Delete deselected spectra from spectra order
        self.spec_order = [
            self.spec_order[i_spec]
            for i_spec in range(len(self.spec_order))
            if not spec_deselected[i_spec]
        ]

        # Infer types, frequencies, and total number of bins
        self.spec_types = [s[:2] for s in self.spec_order]
        self.N_spectra_total = len(self.spec_order)
        self.N_bins_total = jnp.sum(self.crop_mask)
        self.unique_spec_types = list(np.unique(self.spec_types))

        # Crop band powers, covariance matrix, window functions
        self._data_bandpowers = jnp.delete(
            self._data_bandpowers, jnp.invert(self.crop_mask)
        )
        self.data_bandpowers = jnp.delete(
            self.data_bandpowers, jnp.invert(self.crop_mask)
        )
        self.covariance = jnp.delete(
            self.covariance, jnp.invert(self.crop_mask), axis=0
        )
        self.covariance = jnp.delete(
            self.covariance, jnp.invert(self.crop_mask), axis=1
        )

        # Crop window functions
        window_list = []
        for i, wdw in enumerate(self.window_functions):
            if not spec_deselected[i]:
                window_list.append(
                    jnp.delete(
                        wdw,
                        jnp.invert(self.crop_mask)[
                            self.bins_start_ix[i] : self.bins_stop_ix[i]
                        ],
                        axis=1,
                    )
                )
        self.window_functions = window_list

        # Update start and stop indices
        self.bins_start_ix, self.bins_stop_ix = get_start_stop_ix(self.N_bins)

    def generate_crop_mask(self):
        """
        Generate a crop mask. A crop mask is understood to be a boolean mask to be applied to the uncropped data vector
        with True entries for bins that should be used and False entries for bins that should be ignored.
        The method accessed the information on data selection passed by the user interprets it.

        If multiple string hints are passed then only bins selected by the most hints are used.

        See also: interpret_crop_hint

        Returns
        --------------
        array : bool
            The mask to be applied to the uncropped long data vector.
        """

        # Check if crop is necessary
        if not "data_selection" in self.data_set_dict:
            return np.ones(self.N_bins_total) == 1

        if (
            isinstance(self.data_set_dict["data_selection"], str)
            and self.data_set_dict["data_selection"].lower() == "none"
        ):
            return np.ones(self.N_bins_total) == 1

        if self.data_set_dict["data_selection"] is None:
            return np.ones(self.N_bins_total) == 1

        if isinstance(self.data_set_dict["data_selection"], str):
            # Generate crop mask based on some string hint
            crop_mask = self.interpret_crop_hint(self.data_set_dict["data_selection"])
        elif isinstance(self.data_set_dict["data_selection"], list) and isinstance(
            self.data_set_dict["data_selection"][0], str
        ):
            # Generate crop mask based on multiple string hints
            cumulative_crop_mask = np.zeros(self.N_bins_total)
            for this_string_hint in self.data_set_dict["data_selection"]:
                cumulative_crop_mask += self.interpret_crop_hint(this_string_hint)
            crop_mask = np.array(
                cumulative_crop_mask == np.amax(cumulative_crop_mask), dtype=bool
            )
        elif isinstance(self.data_set_dict["data_selection"], list):
            # Can either be series of spectra to use or a boolean mask
            if isinstance(self.data_set_dict["data_selection"][0], str):
                crop_mask = np.array(
                    [
                        1 if spec in self.data_set_dict["data_selection"] else 0
                        for spec in self.spec_order
                    ]
                )
            elif isinstance(self.data_set_dict["data_selection"][0], bool):
                crop_mask = np.array(self.data_set_dict["data_selection"], dtype=int)
            else:
                # Cannot interpret the list passed
                crop_mask = np.ones(self.N_bins_total)
        else:
            crop_mask = np.ones(self.N_bins_total)
        return crop_mask == 1  # cast to bool array

    def interpret_crop_hint(self, crop_hint, verbose=False):
        """
        Interpret a string hint that indicates a crop mask.
        The intended format is: "(data) (action)", where "(data)" specifies which part of the data is selected and
        "(action)" declares what to do with this selection.
        Understood options for "(data)" are:
        * spectrum types, e.g. "pp"
        * ell range, e.g. "ell<650" or "ell>1500"
        Understood options for "(action)" are:
        * "remove" remove this part
        * "only" only keep this part, removing all the rest

        Returns
        --------------
        array : bool
            The mask to be applied to the uncropped long data vector corresponding to this data hint.
        """

        # Split hint into data and action part
        data_hint, action_hint = crop_hint.split(" ")

        affected_specs = []
        base_msk = np.ones(self.N_bins_total) == 0
        if data_hint in self.spec_types:
            # Probably concerns spectra
            for i_spec, st in enumerate(self.spec_types):
                if data_hint == st:
                    base_msk[self.bins_start_ix[i_spec] : self.bins_stop_ix[i_spec]] = (
                        True
                    )
                    affected_specs.append(self.spec_order[i_spec])

        elif "ell" in data_hint:
            # ell cut
            ell_cut = float(data_hint.split("ell")[1][1:])
            compare_operation = data_hint.split("ell")[1][0]
            base_msk[eval(f"self.effective_ells{compare_operation}ell_cut")] = True

        else:
            raise Exception(
                f"Did not recognise data hint: {data_hint}\n Using all bins."
            )
            return np.ones(self.N_bins_total) == 1

        # Go through action hints and check whether we need to invert the selection
        if action_hint == "only":
            pass
        elif action_hint == "remove":
            base_msk = np.invert(base_msk)
        else:
            raise Exception(
                f"Did not recognise action: {action_hint}\n Using all bins."
            )
            return np.ones(self.N_bins_total) == 1

        # Feedback
        if len(affected_specs) > 0:
            specs_to_list = affected_specs
            if action_hint == "remove":
                specs_to_list = [
                    spec for spec in self.spec_order if not spec in affected_specs
                ]
            if verbose:
                print(f"Made a mask that keeps only: {specs_to_list}")

        return base_msk

    def get_ell_helpers(self):
        """
        Calculate useful variables concerning the ell range and binning.

        Returns
        --------------
        array : float
            array running over the angular multipoles for one spectrum (2->N_ell_bins_theory+2)
        array : float
            2d array giving the angular multipole range for each spectrum in rows
        array : float
            long array giving the angular multipole range for each spectrum (concatenated)
        array : float
            long array effective containing the effective bin centres (concatenated according to spectra_order).
        """

        ells = jnp.arange(2, self.N_ell_bins_theory + 2)
        tiled_ells = jnp.tile(ells, (self.N_spectra_total, 1))
        long_ells = tiled_ells.flatten()

        effective_ells = jnp.array([])
        for i, spec in enumerate(self.spec_order):
            effective_ells = jnp.concatenate(
                (
                    effective_ells,
                    jnp.dot(
                        jnp.arange(2, self.N_ell_bins_theory + 2),
                        self.window_functions[i],
                    ).flatten(),
                )
            )

        return ells, tiled_ells, long_ells, effective_ells

    def blind_bandpowers(self, seed=None):
        """
        Applies blinding function to bandpowers by multiplying by a random sinusoid function and a slope.
        It's important that the same spectrum types get hit by the same blinding function, i.e. e.g. the same for all TT spectra regardless of frequency.

        Parameters
        ----------
        seed : int, optional
            random seed to use

        Returns
        -------
        array
            blinded band powers
        array
            blinding function
        """

        rng = np.random.default_rng(seed)

        blind_func_by_spec = {}
        for spec_type in np.unique(self.spec_types):
            ells = None
            for i, spec in enumerate(self.spec_order):
                if spec[:2] == spec_type:
                    ells = self.effective_ells[
                        self.bins_start_ix[i] : self.bins_stop_ix[i]
                    ]
                    break

            blind_func = 1.0
            blind_func += rng.uniform(low=-0.9, high=0.9) * (
                (rng.integers(0, np.amax(ells)) - ells) / np.amax(ells)
            ) + rng.uniform(
                low=-0.1, high=0.1
            )  # tilt
            blind_func += rng.uniform(low=-0.5, high=0.5) * np.sin(
                ells / rng.integers(50, 100) + rng.integers(-100, 100)
            )  # acoustic peaks

            blind_func_by_spec[spec_type] = blind_func
        blind_func = np.block(
            [blind_func_by_spec[spec[:2]] for spec in self.spec_order]
        )

        return jnp.array(self._data_bandpowers * blind_func), jnp.array(blind_func)


# --------------------------------------#
# PRIORS
# --------------------------------------#


class GaussianPrior:
    """
    Base class for Gaussian priors.

    Attributes
    -----------------
    central_value : array, float
        The central value of the prior.
    par_names : list, str
        The names of the parameters this prior acts on.
    prior_covariance : array, float
        The covariance matrix.
    prior_covariance_chol : array, float
        Cholesky decomposition of the covariance matrix.

    Methods
    ----------------
    __init__ :
        initialises an instance of the class.
    log_like :
        Calculates the positive log likelihood of the prior.

    """

    def __init__(self, central_value, prior_covariance, par_names):
        """
        Initialise a new instance of the GaussianPrior class.
        Note that the order of parameters across arguments is expected to be the same, i.e. the [i] central value
        corresponds to the [i,i] entry in the covariance and the [i] parameter name.

        Parameters
        --------------
        central_value : float or array (float)
            Central values of the prior
        prior_covariance : float or array (float)
            Covariance of the prior
        par_names : str or list (str)
            List of names the prior acts on.

        Returns
        --------------
        GaussianPrior :
            A new instance of the GaussianPrior class with the set-up completed.
        """

        self.central_value = jnp.atleast_1d(
            central_value
        )  # Make sure central value is a vector
        self.prior_covariance = jnp.atleast_2d(
            prior_covariance
        )  # Make sure covariance is a matrix
        self.prior_covariance_chol = jnp.linalg.cholesky(self.prior_covariance)

        # par names needs to be a list of strings
        if type(par_names) is str:
            self.par_names = [par_names]
        else:
            self.par_names = par_names

    def log_like(self, sampled_pars):
        """
        Calculates the positive log likelihood of the prior.

        Parameters
        --------------
        sampled_pars : dict
            Dictionary that holds the values of all parameters in par_names.

        Returns
        --------------
        float :
            The positive log likelihood of the prior evaluated for the sampled_params.
        """

        delta_pars = (
            jnp.atleast_1d([sampled_pars[par_name] for par_name in self.par_names])
            - self.central_value
        )
        chol_fac = jnp.linalg.solve(self.prior_covariance_chol, delta_pars)
        logl = (
            jnp.dot(chol_fac.T, chol_fac) / 2
        )  # equivalent to the chisq method, i.e. 0.5*(delta @ C^-1 @ delta)
        return logl


# --------------------------------------#
# HELPER FUNCTIONS
# --------------------------------------#


def get_start_stop_ix(N_bins):
    """
    Generates a list of start anad stop indices given the number of bins in each spectrum.

    Parameters
    --------------
    N_bins : array (int)
        List with an int entry for each spectrum giving the number of bins.

    Returns
    --------------
    array (int) :
        The start indices of each spectrum in a long vector.
    array (int) :
        The stop indices of each spectrum in a long vector.
    """

    bins_stop_ix = np.cumsum(N_bins)
    bins_start_ix = np.insert(bins_stop_ix[:-1], 0, 0)

    return bins_start_ix, bins_stop_ix


def cholesky_decomposition(covariance, N_sims=None, N_bins=None):
    """
    Performs the Ccholesky decomposition of the covariance matrix. Stops the program if unsuccessful.

    Parameters
    --------------
    covariance : array (float)
        The matrix to be decomposed
    N_sims : int, optional
        Number of simulations (used for the Hartlap correction).
    N_bins : int, optional
        Number of bins in the data vector (used for the Hartlap correction).

    Returns
    --------------
    array (float) :
        Cholesky decomposition of the input matrix.
    """

    covariance_chol_dec = None

    # Compute cholesky decomposition, make sure it works
    try:
        covariance_chol_dec = jnp.linalg.cholesky(covariance)
    except:
        raise Exception(
            "Band power covariance matrix is not positive definite! Stopping."
        )
        exit(1)

    # Check no nans are present from insufficient precision
    if np.isnan(covariance_chol_dec).any():
        raise Exception(
            "candl: cholesky decomposition contains 'nan'! Check file and try switching on double precision in JAX. Stopping."
        )
        exit(1)

    # Apply Hartlap correction if requested
    if N_sims is not None and N_bins is not None:
        covariance_chol_dec /= np.sqrt((N_sims - N_bins - 2) / (N_sims - 1))

    return covariance_chol_dec



# --------------------------------------#
# IMPORTS
# --------------------------------------#

from os import listdir
import yaml
import candl
import numpy as np
from copy import deepcopy
import sigfig
import importlib

# --------------------------------------#
# TEST FUNCTIONS
# --------------------------------------#


def run_test(test_file):
    """
    Runs a test based on the provided test file.
    Uses the relative difference between the official likelihood and the candl implementation.
    Relative tolerance is 1e-3.
    Prints results.


    Parameters
    ------------
        test_file : str
            The path to the test .yaml file.

    Returns
    ------------
        None
    """

    # Load info from yaml file
    with open(test_file, "r") as f:
        test_dict = yaml.load(f, Loader=yaml.loader.SafeLoader)

    # Initialise likelihood
    try:
        data_set_str = test_dict["data_set_file"]
        try:
            # Check whether a short cut from a library of the style module.data_set_name is being passed
            module_name, data_set_name = data_set_str.split(".")
            data_module = importlib.import_module(module_name)
            data_set_str = getattr(data_module, data_set_name)
        except:
            pass
        if test_dict["lensing"]:
            candl_like = candl.LensLike(data_set_str, feedback=False)
        else:
            candl_like = candl.Like(data_set_str, feedback=False)
    except:
        print(
            f"!!! -> Failed to initialise likelihood for {test_file} pointing to {data_set_str}"
        )
        return

    # Load test spectrum, juggle into dictionary, grab test parameter values
    try:
        try:
            test_spec = np.loadtxt(
                "/".join(test_file.split("/")[:-1]) + f"/{test_dict['test_spectrum']}"
            )
        except:
            test_spec = np.loadtxt(test_dict["test_spectrum"])
    except:
        print(
            f"!!! -> Failed to load test spectrum for {test_file} pointing to {test_dict['test_spectrum']}"
        )
        return
    pars_for_like = deepcopy(test_dict["param_values"])
    pars_for_like["Dl"] = {}
    test_spec_save_order = ["ell", "TT", "TE", "EE", "BB", "pp", "kk"]
    for i, spec in enumerate(test_spec_save_order):
        pars_for_like["Dl"][spec] = test_spec[:, i]

    # Evaluate likelihood
    rel_diff = 1
    if "test_chisq" in test_dict:
        # Test chisq
        candl_chisq = candl_like.chi_square(pars_for_like)
        rel_diff = (candl_chisq - test_dict["test_chisq"]) / test_dict["test_chisq"]
    elif "test_logl" in test_dict:
        # Test logl
        candl_logl = candl_like.log_like(pars_for_like)
        # Remove prior contribution if desired
        if "remove_prior" in test_dict:
            if test_dict["remove_prior"]:
                candl_logl += candl_like.prior_logl(pars_for_like)
        rel_diff = (candl_logl - test_dict["test_logl"]) / test_dict["test_logl"]

    # Print results
    if abs(rel_diff) < 1e-3:
        print(f"Test passed for {candl_like.name}:")
        print(
            f"  (relative difference = {sigfig.round(float(rel_diff), sigfigs = 2)})!"
        )
    else:
        print(f"!!! -> Test failed for {candl_like.name}:")
        print(
            f"  !!! -> (relative difference = {sigfig.round(float(rel_diff), sigfigs = 2)})!"
        )



"""
Handles imports and deals with optional packages such as JAX.
You can manually set ``USE_JAX = False`` to force numpy/scipy to be used instead of JAX even if it is installed (may be helpful for debugging).
"""

# --------------------------------------#
# NECESSARY
# --------------------------------------#

import os

import inspect

from functools import partial

import importlib

import yaml
import logging

import numpy as np
import scipy as sp

from copy import deepcopy

import sigfig
from textwrap import wrap
from tqdm.notebook import tqdm

from matplotlib.patches import Ellipse
import matplotlib.transforms as transforms
import matplotlib.pyplot as plt
from matplotlib import rc
import seaborn as sns

# --------------------------------------#
# OPTIONAL
# --------------------------------------#

# Global flag that can override use of JAX even if it is installed
# Useful for debugging to print out numpy arrays
USE_JAX = True

# Optional JAX import
JAX_IMPORT_SUCCEEDED = False
try:
    if USE_JAX:
        import jax.numpy as jnp
        import jax.scipy as jsp
        from jax import jit, jacfwd, custom_jvp
        from jax import config as jax_config

        # Unify syntax for setting array elements
        def jax_optional_set_element(arr, ix, el):
            return arr.at[ix].set(el)

        # Need JAX to run with 64bit for dynamic range in covmats
        try:
            jax_config.update("jax_enable_x64", True)
        except:
            raise Exception(
                "candl: could not configure JAX to run in 64 bit mode - this will likely lead to wrong results!"
            )
        JAX_IMPORT_SUCCEEDED = True
    else:
        raise ImportError()
except:
    import numpy as jnp
    import scipy as jsp

    # Unify syntax for setting array elements
    def jax_optional_set_element(arr, ix, el):
        arr[ix] = el
        return arr

    # define jit decorator to do nothing
    def jit(func, **kwargs):
        return func

    # define custom_jvp decorator to do nothing
    def custom_jvp(func, **kwargs):
        func.defjvp = lambda *args: None
        return func

    # define jacfwd to do nothing
    def jacfwd(func, **kwargs):
        return func


# Optional pycapse import
try:
    import pycapse as pc
    import json
except:
    pass

# Optional CAMB import
try:
    import camb
except:
    pass

# Optional CLASS import
try:
    import classy
except:
    pass

# Optional Cobaya import
try:
    from cobaya.theory import Theory as cobaya_theory_Theory
    from cobaya.likelihood import Likelihood as cobaya_likelihood_Likelihood
except:

    # Empty class definitions in case Cobaya is not installed
    class cobaya_theory_Theory:
        pass

    class cobaya_likelihood_Likelihood:
        pass


# Optional CosmoPower import
try:
    import cosmopower as cp
except:
    pass

# Optional CosmoPower-JAX import
try:
    from cosmopower_jax.cosmopower_jax import CosmoPowerJAX as CPJ
except:
    pass

# Optional clipy import
try:
    import clipy
except:
    pass

# Optional candl_data import
try:
    import candl_data
except:
    print(
        "Failed to import `candl_data` (see instructions here for installation: https://github.com/Lbalkenhol/candl_data)."
    )
    print("You can still use likelihoods from other data libraries.")
